import xmltodict
import pandas as pd
import requests
import zipfile
import io
import os
import numpy as np
from pathlib import Path
from datetime import datetime, timedelta, time
from time import time as t

pd.set_option('display.width', None)


class XMLCacheManager:
    """Class for accessing data stored in AEMO's NEMDE output files.

    Examples
    --------

    A XMLCacheManager instance is created by providing the path to directory containing the cache of XML files.

    >>> manager = XMLCacheManager('test_nemde_cache')

    Parameters
    ----------
    cache_folder : str
    """
    def __init__(self, cache_folder):
        self.cache_folder = cache_folder
        self.interval = None
        self.xml = None
        Path(cache_folder).mkdir(parents=False, exist_ok=True)

    def populate(self, start_year, start_month, end_year, end_month, verbose=True):
        """Download data to the cache from the AEMO website. Data downloaded is inclusive of the start and end month."""

        if end_month == 12:
            end_month = 1
            end_year += 1
        else:
            end_month += 1

        self.populate_by_day(start_year=start_year, start_month=start_month, start_day=1,
                             end_year=end_year, end_month=end_month, end_day=1, verbose=verbose)

    def populate_by_day(self, start_year, start_month, end_year, end_month, start_day, end_day, verbose=True):
        """Download data to the cache from the AEMO website. Data downloaded is inclusive of the start and end date."""

        start = datetime(year=start_year, month=start_month, day=start_day) - timedelta(days=1)
        if end_month == 12:
            end_month = 0
            end_year += 1
        end = datetime(year=end_year, month=end_month, day=end_day)
        download_date = start
        while download_date <= end:
            if verbose:
                print('Downloading NEMDE XML file for year={}, month={}, day={}'.format(download_date.year,
                                                                                        download_date.month,
                                                                                        download_date.day))
            download_date_str = download_date.isoformat().replace('T', ' ').replace('-', '/')
            self.load_interval(download_date_str)
            download_date += timedelta(days=1)

    def load_interval(self, interval):
        """Load the data for particular 5 min dispatch interval into memory.

        If the file intervals data is not on disk then an attempt to download it from AEMO's NEMweb portal is made.

        Examples
        --------

        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        Parameters
        ----------
        interval : str
            In the format '%Y/%m/%d %H:%M:%S'

        Raises
        ------
        MissingDataError
            If the data for an interval is not in the cache and cannot be downloaded from NEMWeb.
        """
        self.interval = interval
        if not self.interval_inputs_in_cache():
            self._download_xml_from_nemweb()
            if not self.interval_inputs_in_cache():
                raise MissingDataError('File not downloaded, check internet connection and that NEMWeb contains data for interval {}.'.format(self.interval))
        with open(self.get_file_path()) as file:
            read = file.read()
            self.xml = xmltodict.parse(read)

    def interval_inputs_in_cache(self):
        """Check if the cache contains the data for the loaded interval, primarily for debugging.

        Examples
        --------

        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.interval_inputs_in_cache()
        True

        Returns
        -------
        bool
        """
        return os.path.exists(self.get_file_path())

    def get_file_path(self):
        """Get the file path to the currently loaded interval.

        Examples
        --------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_file_path() # doctest: +SKIP
        PosixPath('test_nemde_cache/NEMSPDOutputs_2018123124000.loaded')

        So the doctest runs on all Operating systems lets also look at the parts of the path.

        >>> manager.get_file_path().parts
        ('test_nemde_cache', 'NEMSPDOutputs_2018123124000.loaded')
        """
        return Path(self.cache_folder) / self.get_file_name()

    def get_file_name(self):
        """Get the filename of the currently loaded interval.

        Examples
        --------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_file_name()
        'NEMSPDOutputs_2018123124000.loaded'
        """
        year, month, day = self._get_market_year_month_day_as_str()
        interval_number = self._get_interval_number_as_str()
        base_name = "NEMSPDOutputs_{year}{month}{day}{interval_number}00.loaded"
        name = base_name.format(year=year, month=month, day=day, interval_number=interval_number)
        path_name = Path(self.cache_folder) / name
        name_OCD = name.replace('.loaded', '_OCD.loaded')
        path_name_OCD = Path(self.cache_folder) / name_OCD
        if os.path.exists(path_name):
            return name
        elif path_name_OCD:
            return name_OCD
        else:
            return name

    def _download_xml_from_nemweb(self):
        year, month, day = self._get_market_year_month_day_as_str()
        base_url = "https://www.nemweb.com.au/Data_Archive/Wholesale_Electricity/NEMDE/{year}/NEMDE_{year}_{month}/NEMDE_Market_Data/NEMDE_Files/NemSpdOutputs_{year}{month}{day}_loaded.zip"
        url = base_url.format(year=year, month=month, day=day)
        r = requests.get(url)
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(self.cache_folder)

    def _get_market_year_month_day(self):
        date_time = self._get_interval_datetime_object()
        hour = date_time.hour
        minute = date_time.minute
        start_market_day = time(hour=4, minute=5)
        interval_time = time(hour=hour, minute=minute)
        if interval_time < start_market_day:
            date_time_for_date = date_time - timedelta(days=1)
        else:
            date_time_for_date = date_time
        year = date_time_for_date.year
        month = date_time_for_date.month
        day = date_time_for_date.day
        return year, month, day

    def _get_interval_number_as_str(self):
        return str(self._get_interval_number()).zfill(3)

    def _get_market_year_month_day_as_str(self):
        year, month, day = self._get_market_year_month_day()
        year_str = str(year)
        month_str = str(month).zfill(2)
        day_str = str(day).zfill(2)
        return year_str, month_str, day_str

    def _get_interval_number(self):
        year, month, day = self._get_market_year_month_day()
        start_market_day_datetime = datetime(year=year, month=month, day=day, hour=4, minute=5)
        time_since_market_day_started = self._get_interval_datetime_object() - start_market_day_datetime
        intervals_elapsed = time_since_market_day_started / timedelta(minutes=5)
        interval_number = int(intervals_elapsed) + 1
        return interval_number

    def _get_interval_datetime_object(self):
        return datetime.strptime(self.interval, '%Y/%m/%d %H:%M:%S')

    def get_unit_initial_conditions(self):
        """Get the initial conditions of units at the start of the dispatch interval.

        Examples
        --------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_unit_initial_conditions()
                 DUID   INITIALMW  RAMPUPRATE  RAMPDOWNRATE  AGCSTATUS
        0      AGLHAL    0.000000         NaN           NaN        0.0
        1      AGLSOM    0.000000         NaN           NaN        0.0
        2     ANGAST1    0.000000         NaN           NaN        0.0
        3       APD01    0.000000         NaN           NaN        0.0
        4       ARWF1   54.500000         NaN           NaN        0.0
        ..        ...         ...         ...           ...        ...
        283  YARWUN_1  140.360001         NaN           NaN        0.0
        284     YWPS1  366.665833  177.750006    177.750006        1.0
        285     YWPS2  374.686066  190.125003    190.125003        1.0
        286     YWPS3    0.000000  300.374994    300.374994        0.0
        287     YWPS4  368.139252  182.249994    182.249994        1.0
        <BLANKLINE>
        [288 rows x 5 columns]

        Returns
        -------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            DUID              unique identifier of a dispatch unit, \n
                              (as `str`)
            INITIALMW         the output or consumption of the unit \n
                              at the start of the interval, in MW, \n
                              (as `np.int64`),
            RAMPUPRATE        ramp up rate of unit as repoted by the \n
                              scada system at the start if the \n
                              interval, in MW/h, (as `np.int64`)
            RAMPDOWNRATE      ramp down rate of unit as repoted by the \n
                              scada system at the start if the \n
                              interval, in MW/h, (as `np.int64`)
            AGCSTATUS         flag to indicate whether the unit is \n
                              connected to the AGC system at the \n
                              start of the interval, 0.0 if not and \n
                              1.0 if it is, (as `np.int64`)
            ================  ========================================

        """
        traders = self.xml['NEMSPDCaseFile']['NemSpdInputs']['TraderCollection']['Trader']
        initial_conditions = dict(DUID=[], INITIALMW=[], RAMPUPRATE=[], RAMPDOWNRATE=[], AGCSTATUS=[])
        if self.is_intervention_period():
            INITIALMW_name = 'WhatIfInitialMW'
        else:
            INITIALMW_name = 'InitialMW'
        name_map = dict(INITIALMW=INITIALMW_name, RAMPUPRATE='SCADARampUpRate', RAMPDOWNRATE='SCADARampDnRate',
                        AGCSTATUS='AGCStatus')
        for trader in traders:
            initial_conditions['DUID'].append(trader['@TraderID'])
            initial_cons = trader['TraderInitialConditionCollection']['TraderInitialCondition']
            for our_name, aemo_name in name_map.items():
                for con in initial_cons:
                    if con['@InitialConditionID'] == aemo_name:
                        value = float(con['@Value'])
                        break
                    else:
                        value = np.NAN
                initial_conditions[our_name].append(value)
        initial_conditions = pd.DataFrame(initial_conditions)
        return initial_conditions

    def get_unit_fast_start_parameters(self):
        """Get the unit fast start dispatch inflexibility parameter values.

        Examples
        --------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_unit_fast_start_parameters()
                DUID  MinLoadingMW  CurrentMode  CurrentModeTime  T1  T2  T3  T4
        0     AGLHAL             2            0                0  10   3  10   2
        1     AGLSOM            16            0                0  20   2  35   2
        2   BARCALDN            12            0                0  14   4   1   4
        3   BARRON-1             5            4                1  11   3   1   1
        4   BARRON-2             5            4                1  11   3   1   1
        ..       ...           ...          ...              ...  ..  ..  ..  ..
        69     VPGS5            48            0                0   5   3  15   0
        70     VPGS6            48            0                0   5   3  15   0
        71   W/HOE#1           160            0                0   3   0   0   0
        72   W/HOE#2           160            0                0   3   0   0   0
        73    YABULU            83            0                0   5   6  42   6
        <BLANKLINE>
        [74 rows x 8 columns]

        Returns
        --------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            DUID              unique identifier of a dispatch unit, \n
                              (as `str`)
            MinLoadingMW      :download:`see AEMO doc <../../docs/pdfs/Fast_Start_Unit_Inflexibility_Profile_Model_October_2014.pdf>`, \n
                              in MW, (as `np.int64`)
            CurrentMode       The dispatch mode if the unit at the \n
                              start of the interval, for mode \n
                              definitions :download:`see AEMO doc <../../docs/pdfs/Fast_Start_Unit_Inflexibility_Profile_Model_October_2014.pdf>`,\n
                              (as `np.int64`)
            CurrentModeTime   The time already spent in the current \n
                              mode, in minutes, (as `np.int64`)
            T1                The total length of mode 1, in minutes \n
                              (as `np.int64`)
            T2                The total length of mode 2, in minutes \n
                              (as `np.int64`)
            T3                The total length of mode 1, in minutes, \n
                              (as `np.int64`)
            T4                The total length of mode 4, in minutes, \n
                              (as `np.int64`)
            ================  ========================================


        """
        traders = self.xml['NEMSPDCaseFile']['NemSpdInputs']['TraderCollection']['Trader']
        initial_conditions = dict(DUID=[], MinLoadingMW=[], CurrentMode=[], CurrentModeTime=[], T1=[], T2=[],
                                  T3=[], T4=[])
        cols = dict(MinLoadingMW='@MinLoadingMW', CurrentMode='@CurrentMode',
                    CurrentModeTime='@CurrentModeTime', T1='@T1', T2='@T2', T3='@T3', T4='@T4')
        for trader in traders:
            row = False
            if '@CurrentMode' in trader:
                for key, name in cols.items():
                    row = True
                    value = int(trader[name])
                    if name == '@CurrentMode' and '@WhatIfCurrentMode' in trader:
                        value = int(trader['@WhatIfCurrentMode'])
                    if name == '@CurrentModeTime' and '@WhatIfCurrentModeTime' in trader:
                        value = int(trader['@WhatIfCurrentModeTime'])
                    initial_conditions[key].append(value)
            if row:
                initial_conditions['DUID'].append(trader['@TraderID'])

        initial_conditions = pd.DataFrame(initial_conditions)
        return initial_conditions

    def get_unit_volume_bids(self):
        """Get the unit volume bids

        Examples
        --------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_unit_volume_bids()
                 DUID     BIDTYPE  MAXAVAIL  ENABLEMENTMIN  ENABLEMENTMAX  LOWBREAKPOINT  HIGHBREAKPOINT  BANDAVAIL1  BANDAVAIL2  BANDAVAIL3  BANDAVAIL4  BANDAVAIL5  BANDAVAIL6  BANDAVAIL7  BANDAVAIL8  BANDAVAIL9  BANDAVAIL10  RAMPDOWNRATE  RAMPUPRATE
        0      AGLHAL      ENERGY     173.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0         0.0        60.0         0.0         0.0        160.0         720.0       720.0
        1      AGLSOM      ENERGY     160.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0        170.0         480.0       480.0
        2     ANGAST1      ENERGY      43.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0        50.0         0.0         0.0         0.0         50.0         840.0       840.0
        3       APD01   LOWER5MIN       0.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0        300.0           0.0         0.0
        4       APD01  LOWER60SEC       0.0            0.0            0.0            0.0             0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0        300.0           0.0         0.0
        ...       ...         ...       ...            ...            ...            ...             ...         ...         ...         ...         ...         ...         ...         ...         ...         ...          ...           ...         ...
        1021    YWPS4   LOWER6SEC      25.0          250.0          385.0          275.0           385.0        15.0        10.0         0.0         0.0         0.0         0.0         0.0         0.0         0.0          0.0           0.0         0.0
        1022    YWPS4   RAISE5MIN       0.0          250.0          390.0          250.0           380.0         0.0         0.0         0.0         0.0         5.0         0.0         0.0         5.0         0.0         10.0           0.0         0.0
        1023    YWPS4    RAISEREG      15.0          250.0          385.0          250.0           370.0         0.0         0.0         0.0         0.0         0.0         0.0         5.0        10.0         0.0          5.0           0.0         0.0
        1024    YWPS4  RAISE60SEC      10.0          220.0          400.0          220.0           390.0         0.0         0.0         0.0         0.0         0.0         5.0         5.0         0.0         0.0         10.0           0.0         0.0
        1025    YWPS4   RAISE6SEC      15.0          220.0          405.0          220.0           390.0         0.0         0.0         0.0        10.0         5.0         0.0         0.0         0.0         0.0         10.0           0.0         0.0
        <BLANKLINE>
        [1026 rows x 19 columns]

        Returns
        --------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            DUID              unique identifier of a dispatch unit, \n
                              (as `str`)
            BIDTYPE           the service the bid applies to, \n
                              (as `str`)
            MAXAVAIL          the bid in unit availablity, in MW,
                              (as `str`)
            ENABLEMENTMIN     :download:`see AMEO docs  <../../docs/pdfs/FCAS Model in NEMDE.pdf>`, \n
                              in MW, (as `np.float64`)
            ENABLEMENTMAX     :download:`see AMEO docs  <../../docs/pdfs/FCAS Model in NEMDE.pdf>`, \n
                              in MW, (as `np.float64`)
            LOWBREAKPOINT     :download:`see AMEO docs <../../docs/pdfs/FCAS Model in NEMDE.pdf>`, \n
                              in MW, (as `np.float64`)
            HIGHBREAKPOINT    :download:`see AMEO docs  <../../docs/pdfs/FCAS Model in NEMDE.pdf>`, \n
                              in MW, (as `np.float64`)
            BANDAVAIL1        the volume bid in the first bid band,
                              in MW, (as `np.float64`)
                 :
            BANDAVAIL10       the volume bid in the tenth bid band,
                              in MW, (as `np.float64`)
            RAMPDOWNRATE      the bid in ramp down rate, in MW/h,
                              (as `np.int64`)
            RAMPUPRATE        the bid in ramp up rate, in MW/h,
                              (as `np.int64`)
            ================  ========================================


        """
        traders = self.xml['NEMSPDCaseFile']['NemSpdInputs']['PeriodCollection']['Period']['TraderPeriodCollection']['TraderPeriod']
        trades_by_unit_and_type = dict(DUID=[], BIDTYPE=[], MAXAVAIL=[], ENABLEMENTMIN=[], ENABLEMENTMAX=[],
                                       LOWBREAKPOINT=[], HIGHBREAKPOINT=[], BANDAVAIL1=[], BANDAVAIL2=[],
                                       BANDAVAIL3=[], BANDAVAIL4=[], BANDAVAIL5=[], BANDAVAIL6=[], BANDAVAIL7=[],
                                       BANDAVAIL8=[], BANDAVAIL9=[], BANDAVAIL10=[], RAMPDOWNRATE=[], RAMPUPRATE=[])
        name_map = dict(BIDTYPE='@TradeType', MAXAVAIL='@MaxAvail', ENABLEMENTMIN='@EnablementMin',
                        ENABLEMENTMAX='@EnablementMax',  LOWBREAKPOINT='@LowBreakpoint',
                        HIGHBREAKPOINT='@HighBreakpoint', BANDAVAIL1='@BandAvail1', BANDAVAIL2='@BandAvail2',
                        BANDAVAIL3='@BandAvail3', BANDAVAIL4='@BandAvail4', BANDAVAIL5='@BandAvail5',
                        BANDAVAIL6='@BandAvail6', BANDAVAIL7='@BandAvail7', BANDAVAIL8='@BandAvail8',
                        BANDAVAIL9='@BandAvail9', BANDAVAIL10='@BandAvail10', RAMPDOWNRATE='@RampDnRate',
                        RAMPUPRATE='@RampUpRate')
        for trader in traders:
            if type(trader['TradeCollection']['Trade']) != list:
                trades = trader['TradeCollection']
                for _, trade in trades.items():
                    trades_by_unit_and_type['DUID'].append(trader['@TraderID'])
                    for our_name, aemo_name in name_map.items():
                        if aemo_name in trade:
                            if aemo_name == '@TradeType':
                                value = trade[aemo_name]
                            else:
                                value = float(trade[aemo_name])
                        else:
                            value = 0.0
                        trades_by_unit_and_type[our_name].append(value)
            else:
                for trade in trader['TradeCollection']['Trade']:
                    trades_by_unit_and_type['DUID'].append(trader['@TraderID'])
                    for our_name, aemo_name in name_map.items():
                        if aemo_name in trade:
                            if aemo_name == '@TradeType':
                                value = trade[aemo_name]
                            else:
                                value = float(trade[aemo_name])
                        else:
                            value = 0.0
                        trades_by_unit_and_type[our_name].append(value)
        trades_by_unit_and_type = pd.DataFrame(trades_by_unit_and_type)
        bid_type_map = dict(ENOF='ENERGY', LDOF='ENERGY', DROF='ENERGY', L5RE='LOWERREG', R5RE='RAISEREG',
                            R5MI='RAISE5MIN', L5MI='LOWER5MIN', R60S='RAISE60SEC', L60S='LOWER60SEC', R6SE='RAISE6SEC',
                            L6SE='LOWER6SEC')
        trades_by_unit_and_type["BIDTYPE"] = trades_by_unit_and_type["BIDTYPE"].apply(lambda x: bid_type_map[x])
        return trades_by_unit_and_type

    def get_unit_price_bids(self):
        """Get the unit volume bids

        Examples
        --------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_unit_price_bids()
                 DUID     BIDTYPE  PRICEBAND1  PRICEBAND2  PRICEBAND3  PRICEBAND4  PRICEBAND5  PRICEBAND6  PRICEBAND7  PRICEBAND8  PRICEBAND9  PRICEBAND10
        0      AGLHAL      ENERGY    -1000.00        0.00      278.81      368.81      418.81      498.81      578.81     1365.56    10578.87     13998.99
        1      AGLSOM      ENERGY    -1000.00        0.00       85.00      110.00      145.00      284.00      451.00     1001.00    13300.87     14499.96
        2     ANGAST1      ENERGY    -1000.00        0.00      125.00      200.20      299.19      379.98      589.99     1374.85    10618.00     14500.00
        3       APD01   LOWER5MIN        0.00        1.00        2.00        3.00        4.00        5.00        6.00        7.00        8.00         9.00
        4       APD01  LOWER60SEC        0.00        1.00        2.00        3.00        4.00        5.00        6.00        7.00        8.00         9.00
        ...       ...         ...         ...         ...         ...         ...         ...         ...         ...         ...         ...          ...
        1021    YWPS4   LOWER6SEC        0.03        0.05        0.16        0.30        1.90       25.04       30.04       99.00     4600.00      9899.00
        1022    YWPS4   RAISE5MIN        0.05        1.78        4.48       14.50       30.03       49.00       87.70      100.00    11990.00     12400.40
        1023    YWPS4    RAISEREG        0.05        2.70        9.99       19.99       49.00       95.50      240.00      450.50      950.50     11900.00
        1024    YWPS4  RAISE60SEC        0.17        1.80        4.80       10.01       21.00       39.00       52.00      102.00     4400.00     11999.00
        1025    YWPS4   RAISE6SEC        0.48        1.75        4.90       20.70       33.33       99.90      630.00     1999.00     6000.00     12299.00
        <BLANKLINE>
        [1026 rows x 12 columns]


        Returns
        --------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            DUID              unique identifier of a dispatch unit, \n
                              (as `str`)
            BIDTYPE           the service the bid applies to, \n
                              (as `str`)
            PRICEBAND1        the volume bid in the first bid band,
                              in MW, (as `np.float64`)
                 :
            PRICEBAND10       the volume bid in the tenth bid band,
                              in MW, (as `np.float64`)
            ================  ========================================


        """
        traders = self.xml['NEMSPDCaseFile']['NemSpdInputs']['TraderCollection']['Trader']
        trades_by_unit_and_type = dict(DUID=[], BIDTYPE=[], PRICEBAND1=[], PRICEBAND2=[], PRICEBAND3=[], PRICEBAND4=[],
                                       PRICEBAND5=[], PRICEBAND6=[], PRICEBAND7=[], PRICEBAND8=[], PRICEBAND9=[],
                                       PRICEBAND10=[])
        name_map = dict(BIDTYPE='@TradeType', PRICEBAND1='@PriceBand1', PRICEBAND2='@PriceBand2',
                        PRICEBAND3='@PriceBand3', PRICEBAND4='@PriceBand4', PRICEBAND5='@PriceBand5',
                        PRICEBAND6='@PriceBand6', PRICEBAND7='@PriceBand7', PRICEBAND8='@PriceBand8',
                        PRICEBAND9='@PriceBand9', PRICEBAND10='@PriceBand10')
        for trader in traders:
            if type(trader['TradePriceStructureCollection']['TradePriceStructure']['TradeTypePriceStructureCollection']
                    ['TradeTypePriceStructure']) != list:
                trades = trader['TradePriceStructureCollection']['TradePriceStructure']['TradeTypePriceStructureCollection']
                for _, trade in trades.items():
                    trades_by_unit_and_type['DUID'].append(trader['@TraderID'])
                    for our_name, aemo_name in name_map.items():
                        if aemo_name in trade:
                            if aemo_name == '@TradeType':
                                value = trade[aemo_name]
                            else:
                                value = float(trade[aemo_name])
                        else:
                            value = 0.0
                        trades_by_unit_and_type[our_name].append(value)
            else:
                for trade in trader['TradePriceStructureCollection']['TradePriceStructure']['TradeTypePriceStructureCollection']['TradeTypePriceStructure']:
                    trades_by_unit_and_type['DUID'].append(trader['@TraderID'])
                    for our_name, aemo_name in name_map.items():
                        if aemo_name in trade:
                            if aemo_name == '@TradeType':
                                value = trade[aemo_name]
                            else:
                                value = float(trade[aemo_name])
                        else:
                            value = 0.0
                        trades_by_unit_and_type[our_name].append(value)
        trades_by_unit_and_type = pd.DataFrame(trades_by_unit_and_type)
        bid_type_map = dict(ENOF='ENERGY', LDOF='ENERGY', DROF='ENERGY', L5RE='LOWERREG', R5RE='RAISEREG',
                            R5MI='RAISE5MIN', L5MI='LOWER5MIN', R60S='RAISE60SEC', L60S='LOWER60SEC', R6SE='RAISE6SEC',
                            L6SE='LOWER6SEC')
        trades_by_unit_and_type["BIDTYPE"] = trades_by_unit_and_type["BIDTYPE"].apply(lambda x: bid_type_map[x])
        return trades_by_unit_and_type

    def get_UIGF_values(self):
        """Get the unit unconstrained intermittent generation forecast.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_UIGF_values()
                DUID    UIGF
        0      ARWF1  56.755
        1   BALDHWF1   9.160
        2      BANN1   0.000
        3     BLUFF1   4.833
        4     BNGSF1   0.000
        ..       ...     ...
        57     WGWF1  25.445
        58   WHITSF1   0.000
        59  WOODLWN1   0.075
        60     WRSF1   0.000
        61     WRWF1  15.760
        <BLANKLINE>
        [62 rows x 2 columns]

        Returns
        -------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            DUID              unique identifier of a dispatch unit, \n
                              (as `str`)
            UGIF              the units generation forecast for end \n
                              of the inteval, in MW, (as `np.float64`)
            ================  ========================================


        """
        traders = self.xml['NEMSPDCaseFile']['NemSpdInputs']['PeriodCollection']['Period']['TraderPeriodCollection']['TraderPeriod']
        trades_by_unit_and_type = dict(DUID=[], UIGF=[])
        for trader in traders:
            if '@UIGF' in trader:
                trades_by_unit_and_type['DUID'].append(trader['@TraderID'])
                trades_by_unit_and_type['UIGF'].append(float(trader['@UIGF']))
        trades_by_unit_and_type = pd.DataFrame(trades_by_unit_and_type)
        return trades_by_unit_and_type

    def get_violations(self):
        """Get the total volume violation of different constraint sets.

        For more information on the constraint sets :download:`see AMEO docs  <../../docs/pdfs/Schedule of Constraint Violation Penalty factors.pdf>`

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_violations()
        {'regional_demand': 0.0, 'interocnnector': 0.0, 'generic_constraint': 0.0, 'ramp_rate': 0.0, 'unit_capacity': 0.36, 'energy_constraint': 0.0, 'energy_offer': 0.0, 'fcas_profile': 0.0, 'fast_start': 0.0, 'mnsp_ramp_rate': 0.0, 'msnp_offer': 0.0, 'mnsp_capacity': 0.0, 'ugif': 0.0}

        Returns
        -------
        dict

        """
        outputs = self.xml['NEMSPDCaseFile']['NemSpdOutputs']
        name_map = dict(regional_demand='@TotalAreaGenViolation',
                        interocnnector='@TotalInterconnectorViolation',
                        generic_constraint='@TotalGenericViolation',
                        ramp_rate='@TotalRampRateViolation',
                        unit_capacity='@TotalUnitMWCapacityViolation',
                        energy_constraint='@TotalEnergyConstrViolation',
                        energy_offer='@TotalEnergyOfferViolation',
                        fcas_profile='@TotalASProfileViolation',
                        fast_start='@TotalFastStartViolation',
                        mnsp_ramp_rate='@TotalMNSPRampRateViolation',
                        msnp_offer='@TotalMNSPOfferViolation',
                        mnsp_capacity='@TotalMNSPCapacityViolation',
                        ugif='@TotalUIGFViolation')
        violations = {}
        if type(outputs['PeriodSolution']) == list:
            for solution in outputs['PeriodSolution']:
                if solution['@Intervention'] == '0':
                    for name, aemo_name in name_map.items():
                        violations[name] = float(solution[aemo_name])
        else:
            for name, aemo_name in name_map.items():
                violations[name] = float(outputs['PeriodSolution'][aemo_name])
        return violations

    def get_constraint_violation_prices(self):
        """Get the price of violating different constraint sets.

        For more information on the constraint sets :download:`see AMEO docs  <../../docs/pdfs/Schedule of Constraint Violation Penalty factors.pdf>`

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_constraint_violation_prices()
        {'regional_demand': 2175000.0, 'interocnnector': 16675000.0, 'generic_constraint': 435000.0, 'ramp_rate': 16747500.0, 'unit_capacity': 5365000.0, 'energy_offer': 16457500.0, 'fcas_profile': 2247500.0, 'fcas_max_avail': 2247500.0, 'fcas_enablement_min': 1015000.0, 'fcas_enablement_max': 1015000.0, 'fast_start': 16385000.0, 'mnsp_ramp_rate': 16747500.0, 'msnp_offer': 16457500.0, 'mnsp_capacity': 5292500.0, 'uigf': 5582500.0, 'voll': 14500.0}

        Returns
        -------
        dict
        """
        inputs = self.xml['NEMSPDCaseFile']['NemSpdInputs']
        name_map = dict(regional_demand='@EnergyDeficitPrice',
                        interocnnector='@InterconnectorPrice',
                        generic_constraint='@GenericConstraintPrice',
                        ramp_rate='@RampRatePrice',
                        unit_capacity='@CapacityPrice',
                        energy_offer='@OfferPrice',
                        fcas_profile='@ASProfilePrice',
                        fcas_max_avail='@ASMaxAvailPrice',
                        fcas_enablement_min='@ASEnablementMinPrice',
                        fcas_enablement_max='@ASEnablementMaxPrice',
                        fast_start='@FastStartPrice',
                        mnsp_ramp_rate='@MNSPRampRatePrice',
                        msnp_offer='@MNSPOfferPrice',
                        mnsp_capacity='@MNSPCapacityPrice',
                        uigf='@UIGFSurplusPrice',
                        voll='@VoLL')
        violations = {}
        for name, aemo_name in name_map.items():
            violations[name] = float(inputs['Case'][aemo_name])
        return violations

    def is_intervention_period(self):
        """Check if the interval currently loaded was subject to an intervention.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.is_intervention_period()
        False

        Returns
        -------
        bool

        """
        return type(self.xml['NEMSPDCaseFile']['NemSpdOutputs']['PeriodSolution']) == list

    def get_constraint_rhs(self):
        """Get generic constraints rhs values.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_constraint_rhs()
                             set           rhs
        0               #BANN1_E     32.000000
        1              #BNGSF2_E      3.000000
        2            #CROWLWF1_E     43.000000
        3             #CSPVPS1_E     29.000000
        4             #DAYDSF1_E      0.000000
        ..                   ...           ...
        704          V_OWF_NRB_0  10000.001000
        705  V_OWF_TGTSNRBHTN_30  10030.000000
        706        V_S_NIL_ROCOF    812.280029
        707          V_T_NIL_BL1    478.000000
        708        V_T_NIL_FCSPS    425.154024
        <BLANKLINE>
        [709 rows x 2 columns]

        Returns
        -------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            set               the unique identifier of the generic \n
                              constraint, (as `str`)
            rhs               the rhs value of the constraint, \n
                              (as `np.float64`)
            ================  ========================================

        """
        constraints = self.xml['NEMSPDCaseFile']['NemSpdOutputs']['ConstraintSolution']
        rhs_values = dict(set=[], rhs=[])
        for con in constraints:
            if con['@Intervention'] == '0':
                rhs_values['set'].append(con['@ConstraintID'])
                rhs_values['rhs'].append(float(con['@RHS']))
        return pd.DataFrame(rhs_values)

    def get_constraint_type(self):
        """Get generic constraints type.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_constraint_type()
                             set type       cost
        0               #BANN1_E   LE  5220000.0
        1              #BNGSF2_E   LE  5220000.0
        2            #CROWLWF1_E   LE  5220000.0
        3             #CSPVPS1_E   LE  5220000.0
        4             #DAYDSF1_E   LE  5220000.0
        ..                   ...  ...        ...
        704          V_OWF_NRB_0   LE  5220000.0
        705  V_OWF_TGTSNRBHTN_30   LE  5220000.0
        706        V_S_NIL_ROCOF   LE   507500.0
        707          V_T_NIL_BL1   LE  5220000.0
        708        V_T_NIL_FCSPS   LE   435000.0
        <BLANKLINE>
        [709 rows x 3 columns]

        Returns
        -------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            set               the unique identifier of the generic \n
                              constraint, (as `str`)
            type              the type of constraint, i.e '=', '<=' or \n
                              '<=', (as `str`)
            cost              the cost of violating the constraint, \n
                              (as `np.float64`)
            ================  ========================================
        """

        constraints = self.xml['NEMSPDCaseFile']['NemSpdInputs']['GenericConstraintCollection']['GenericConstraint']
        rhs_values = dict(set=[], type=[], cost=[])
        for con in constraints:
            rhs_values['set'].append(con['@ConstraintID'])
            rhs_values['type'].append(con['@Type'])
            rhs_values['cost'].append(float(con['@ViolationPrice']))
        return pd.DataFrame(rhs_values)

    def get_constraint_region_lhs(self):
        """Get generic constraints lhs term regional coefficients.

        This is a compact way of describing constraints that apply to all units in a region. If a constraint set appears
        here and also in the unit specific lhs table then the coefficents used in the constraint is the sum of the two
        coefficients, this can be used to exclude particular units from otherwise region wide constraints.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_constraint_region_lhs()
                           set region service  coefficient
        0        F_I+LREG_0120   NSW1    L5RE          1.0
        1        F_I+LREG_0120   QLD1    L5RE          1.0
        2        F_I+LREG_0120    SA1    L5RE          1.0
        3        F_I+LREG_0120   TAS1    L5RE          1.0
        4        F_I+LREG_0120   VIC1    L5RE          1.0
        ..                 ...    ...     ...          ...
        478   F_T+NIL_WF_TG_R5   TAS1    R5RE          1.0
        479   F_T+NIL_WF_TG_R6   TAS1    R6SE          1.0
        480  F_T+NIL_WF_TG_R60   TAS1    R60S          1.0
        481      F_T+RREG_0050   TAS1    R5RE          1.0
        482    F_T_NIL_MINP_R6   TAS1    R6SE          1.0
        <BLANKLINE>
        [483 rows x 4 columns]

        Returns
        -------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            set               the unique identifier of the generic \n
                              constraint, (as `str`)
            region            the regions the constraint applies in, \n
                              (as `str`)
            service           the services the constraint applies too, \n
                              (as `str`)
            coefficient       the coefficient of the terms on the lhs, \n
                              (as `np.float64`)
            ================  ========================================
        """
        constraints = self.xml['NEMSPDCaseFile']['NemSpdInputs']['GenericConstraintCollection']['GenericConstraint']
        lhs_values = dict(set=[], region=[], service=[], coefficient=[])
        for con in constraints:
            lhs = con['LHSFactorCollection']
            if lhs is not None and 'RegionFactor' in lhs:
                if type(lhs['RegionFactor']) == list:
                    for term in lhs['RegionFactor']:
                        lhs_values['set'].append(con['@ConstraintID'])
                        lhs_values['region'].append(term['@RegionID'])
                        lhs_values['service'].append(term['@TradeType'])
                        lhs_values['coefficient'].append(float(term['@Factor']))
                else:
                    term = lhs['RegionFactor']
                    lhs_values['set'].append(con['@ConstraintID'])
                    lhs_values['region'].append(term['@RegionID'])
                    lhs_values['service'].append(term['@TradeType'])
                    lhs_values['coefficient'].append(float(term['@Factor']))
        return pd.DataFrame(lhs_values)

    def get_constraint_unit_lhs(self):
        """Get generic constraints lhs term unit coefficients.

        If a constraint set appears here and also in the region lhs table then the coefficents used in the
        constraint is the sum of the two coefficients, this can be used to exclude particular units from otherwise
        region wide constraints.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_constraint_unit_lhs()
                              set      unit service  coefficient
        0                #BANN1_E     BANN1    ENOF          1.0
        1               #BNGSF2_E    BNGSF2    ENOF          1.0
        2             #CROWLWF1_E  CROWLWF1    ENOF          1.0
        3              #CSPVPS1_E   CSPVPS1    ENOF          1.0
        4              #DAYDSF1_E   DAYDSF1    ENOF          1.0
        ...                   ...       ...     ...          ...
        5864      V_ARWF_FSTTRP_5     ARWF1    ENOF          1.0
        5865      V_MTGBRAND_33WT  MTGELWF1    ENOF          1.0
        5866     V_OAKHILL_TFB_42  OAKLAND1    ENOF          1.0
        5867          V_OWF_NRB_0  OAKLAND1    ENOF          1.0
        5868  V_OWF_TGTSNRBHTN_30  OAKLAND1    ENOF          1.0
        <BLANKLINE>
        [5869 rows x 4 columns]

        Returns
        -------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            set               the unique identifier of the generic \n
                              constraint, (as `str`)
            unit              the units the constraint applies in, \n
                              (as `str`)
            service           the services the constraint applies too, \n
                              (as `str`)
            coefficient       the coefficient of the terms on the lhs, \n
                              (as `np.float64`)
            ================  ========================================
        """
        constraints = self.xml['NEMSPDCaseFile']['NemSpdInputs']['GenericConstraintCollection']['GenericConstraint']
        lhs_values = dict(set=[], unit=[], service=[], coefficient=[])
        for con in constraints:
            lhs = con['LHSFactorCollection']
            if lhs is not None and 'TraderFactor' in lhs:
                if type(lhs['TraderFactor']) == list:
                    for term in lhs['TraderFactor']:
                        lhs_values['set'].append(con['@ConstraintID'])
                        lhs_values['unit'].append(term['@TraderID'])
                        lhs_values['service'].append(term['@TradeType'])
                        lhs_values['coefficient'].append(float(term['@Factor']))
                else:
                    term = lhs['TraderFactor']
                    lhs_values['set'].append(con['@ConstraintID'])
                    lhs_values['unit'].append(term['@TraderID'])
                    lhs_values['service'].append(term['@TradeType'])
                    lhs_values['coefficient'].append(float(term['@Factor']))
        return pd.DataFrame(lhs_values)

    def get_constraint_interconnector_lhs(self):
        """Get generic constraints lhs term interconnector coefficients.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_constraint_interconnector_lhs()
                             set interconnector  coefficient
        0               DATASNAP      N-Q-MNSP1          1.0
        1        DATASNAP_DFS_LS      N-Q-MNSP1          1.0
        2      DATASNAP_DFS_NCAN      N-Q-MNSP1          1.0
        3    DATASNAP_DFS_NCWEST      N-Q-MNSP1          1.0
        4      DATASNAP_DFS_NNTH      N-Q-MNSP1          1.0
        ..                   ...            ...          ...
        631      V^^S_NIL_TBSE_1           V-SA          1.0
        632      V^^S_NIL_TBSE_2           V-SA          1.0
        633        V_S_NIL_ROCOF           V-SA          1.0
        634          V_T_NIL_BL1      T-V-MNSP1         -1.0
        635        V_T_NIL_FCSPS      T-V-MNSP1         -1.0
        <BLANKLINE>
        [636 rows x 3 columns]

        Returns
        -------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            set               the unique identifier of the generic \n
                              constraint, (as `str`)
            interconnector    the interconnector the constraint applies in, \n
                              (as `str`)
            coefficient       the coefficient of the terms on the lhs, \n
                              (as `np.float64`)
            ================  ========================================
        """
        constraints = self.xml['NEMSPDCaseFile']['NemSpdInputs']['GenericConstraintCollection']['GenericConstraint']
        lhs_values = dict(set=[], interconnector=[], coefficient=[])
        for con in constraints:
            lhs = con['LHSFactorCollection']
            if lhs is not None and 'InterconnectorFactor' in lhs:
                if type(lhs['InterconnectorFactor']) == list:
                    for term in lhs['InterconnectorFactor']:
                        lhs_values['set'].append(con['@ConstraintID'])
                        lhs_values['interconnector'].append(term['@InterconnectorID'])
                        lhs_values['coefficient'].append(float(term['@Factor']))
                else:
                    term = lhs['InterconnectorFactor']
                    lhs_values['set'].append(con['@ConstraintID'])
                    lhs_values['interconnector'].append(term['@InterconnectorID'])
                    lhs_values['coefficient'].append(float(term['@Factor']))
        return pd.DataFrame(lhs_values)

    def get_market_interconnector_link_bid_availability(self):
        """Get the bid availability of market interconnectors.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.get_market_interconnector_link_bid_availability()
          interconnector to_region  availability
        0      T-V-MNSP1      TAS1         478.0
        1      T-V-MNSP1      VIC1         478.0

        Returns
        -------
        pd.DataFrame

            ================  ========================================
            Columns:          Description:
            interconnector    the interconnector the constraint applies in, \n
                              (as `str`)
            to_region         the direction the bid availability applies to, \n
                              (as `str`)
            availability      the availability as bid in by the \n
                              interconnector, (as `str`)
            ================  ========================================
        """
        inters = self.xml['NEMSPDCaseFile']['NemSpdInputs']['PeriodCollection']['Period']['InterconnectorPeriodCollection']['InterconnectorPeriod']
        bid_availability = dict(interconnector=[], to_region=[], availability=[])
        for inter in inters:
            if inter['@MNSP'] == '1':
                for offer in inter['MNSPOfferCollection']['MNSPOffer']:
                    bid_availability['interconnector'].append(inter['@InterconnectorID'])
                    bid_availability['to_region'].append(offer['@RegionID'])
                    bid_availability['availability'].append(float(offer['@MaxAvail']))
        bid_availability = pd.DataFrame(bid_availability)
        return bid_availability

    def find_intervals_with_violations(self, limit, start_year, start_month, end_year, end_month):
        """Find the set of dispatch intervals where the non-intervention dispatch runs had constraint violations.

        Examples
        -------
        >>> manager = XMLCacheManager('test_nemde_cache')

        >>> manager.load_interval('2019/01/01 00:00:00')

        >>> manager.find_intervals_with_violations(limit=3, start_year=2019, start_month=1, end_year=2019, end_month=1)
        {'2019/01/01 00:00:00': ['unit_capacity'], '2019/01/01 00:05:00': ['unit_capacity'], '2019/01/01 00:10:00': ['unit_capacity']}

        Parameters
        ----------
        limit : int
            number of intervals to find, finds first intervals in chronolgical order
        start_year : int
            year to start search
        start_month : int
            month to start search
        end_year : int
            year to end search
        end_month : int
            month to end search

        Returns
        -------
        dict

        """
        if end_month == 12:
            start = datetime(year=start_year, month=start_month, day=1)
            end = datetime(year=end_year + 1, month=1, day=1)
        else:
            start = datetime(year=start_year, month=start_month, day=1)
            end = datetime(year=end_year, month=end_month + 1, day=1)

        check_time = start
        intervals = {}
        while check_time <= end and len(intervals) < limit:
            time_as_str = check_time.isoformat().replace('T', ' ').replace('-', '/')
            try:
                self.load_interval(time_as_str)
                violations = self.get_violations()
                for violation_type, violation_value in violations.items():
                    if violation_value > 0.0:
                        if time_as_str not in intervals:
                            intervals[time_as_str] = []
                        intervals[time_as_str].append(violation_type)
            except MissingDataError:
                pass
            check_time += timedelta(minutes=5)
        return intervals


class MissingDataError(Exception):
    """Raise for unable to downloaded data from NEMWeb."""



import numpy as np
import pandas as pd
from mip import Model, xsum, minimize, CONTINUOUS, OptimizationStatus, BINARY, CBC, GUROBI, LP_Method


class InterfaceToSolver:
    """A wrapper for the mip model class, allows interaction with mip using pd.DataFrames."""

    def __init__(self, solver_name='CBC'):
        self.variables = {}
        self.linear_mip_variables = {}

        self.solver_name = solver_name
        if solver_name == 'CBC':
            self.mip_model = Model("market", solver_name=CBC)
            self.linear_mip_model = Model("market", solver_name=CBC)
        elif solver_name == 'GUROBI':
            self.mip_model = Model("market", solver_name=GUROBI)
            self.linear_mip_model = Model("market", solver_name=GUROBI)
        else:
            raise ValueError("Solver '{}' not recognised.")

        self.mip_model.verbose = 0
        self.mip_model.solver.set_mip_gap_abs(1e-10)
        self.mip_model.solver.set_mip_gap(1e-20)
        self.mip_model.lp_method = LP_Method.DUAL

        self.linear_mip_model.verbose = 0
        self.linear_mip_model.solver.set_mip_gap_abs(1e-10)
        self.linear_mip_model.solver.set_mip_gap(1e-20)
        self.linear_mip_model.lp_method = LP_Method.DUAL

    def add_variables(self, decision_variables):
        """Add decision variables to the model.

        Examples
        --------
        >>> decision_variables = pd.DataFrame({
        ...   'variable_id': [0, 1],
        ...   'lower_bound': [0.0, 0.0],
        ...   'upper_bound': [6.0, 1.0],
        ...   'type': ['continuous', 'binary']})

        >>> si = InterfaceToSolver()

        >>> si.add_variables(decision_variables)

        The underlying mip_model should now have 2 variables.

        >>> print(si.mip_model.num_cols)
        2

        The first one should have the following properties.

        >>> print(si.mip_model.var_by_name('0').var_type)
        C

        >>> print(si.mip_model.var_by_name('0').lb)
        0.0

        >>> print(si.mip_model.var_by_name('0').ub)
        6.0

        The second one should have the following properties.

        >>> print(si.mip_model.var_by_name('1').var_type)
        B

        >>> print(si.mip_model.var_by_name('1').lb)
        0.0

        >>> print(si.mip_model.var_by_name('1').ub)
        1.0

        """
        # Create a mapping between the nempy level names for variable types and the mip representation.
        variable_types = {'continuous': CONTINUOUS, 'binary': BINARY}
        # Add each variable to the mip model.
        for variable_id, lower_bound, upper_bound, variable_type in zip(
                list(decision_variables['variable_id']), list(decision_variables['lower_bound']),
                list(decision_variables['upper_bound']), list(decision_variables['type'])):
            self.variables[variable_id] = self.mip_model.add_var(lb=lower_bound, ub=upper_bound,
                                                                 var_type=variable_types[variable_type],
                                                                 name=str(variable_id))

            self.linear_mip_variables[variable_id] = self.linear_mip_model.add_var(lb=lower_bound, ub=upper_bound,
                                                                                   var_type=variable_types[
                                                                                       variable_type],
                                                                                   name=str(variable_id))

    def add_sos_type_2(self, sos_variables, sos_id_columns, position_column):
        """Add groups of special ordered sets of type 2 two the mip model.

        Examples
        --------

        >>> decision_variables = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'lower_bound': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ...   'upper_bound': [5.0, 5.0, 5.0, 5.0, 5.0, 5.0],
        ...   'type': ['continuous', 'continuous', 'continuous',
        ...            'continuous', 'continuous', 'continuous']})

        >>> sos_variables = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'sos_id': ['A', 'A', 'A', 'B', 'B', 'B'],
        ...   'position': [0, 1, 2, 0, 1, 2]})

        >>> si = InterfaceToSolver()

        >>> si.add_variables(decision_variables)

        >>> si.add_sos_type_2(sos_variables, 'sos_id', 'position')

        """

        # Function that adds sets to mip model.
        def add_sos_vars(sos_group):
            self.mip_model.add_sos(list(zip(sos_group['vars'], sos_group[position_column])), 2)

        # For each variable_id get the variable object from the mip model
        sos_variables['vars'] = sos_variables['variable_id'].apply(lambda x: self.variables[x])
        # Break up the sets based on their id and add them to the model separately.
        sos_variables.groupby(sos_id_columns).apply(add_sos_vars)
        # This is a hack to make sure mip knows there are binary constraints.
        self.mip_model.add_var(var_type=BINARY, obj=0.0)

    def add_sos_type_1(self, sos_variables):
        # Function that adds sets to mip model.
        def add_sos_vars(sos_group):
            self.mip_model.add_sos(list(zip(sos_group['vars'], [1.0 for i in range(len(sos_variables['vars']))])), 1)

        # For each variable_id get the variable object from the mip model
        sos_variables['vars'] = sos_variables['variable_id'].apply(lambda x: self.variables[x])
        # Break up the sets based on their id and add them to the model separately.
        sos_variables.groupby('sos_id').apply(add_sos_vars)
        # This is a hack to make mip knows there are binary constraints.
        self.mip_model.add_var(var_type=BINARY, obj=0.0)

    def add_objective_function(self, objective_function):
        """Add the objective function to the mip model.

        Examples
        --------

        >>> decision_variables = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'lower_bound': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ...   'upper_bound': [5.0, 5.0, 5.0, 5.0, 5.0, 5.0],
        ...   'type': ['continuous', 'continuous', 'continuous',
        ...            'continuous', 'continuous', 'continuous']})

        >>> objective_function = pd.DataFrame({
        ...   'variable_id': [0, 1, 3, 4, 5],
        ...   'cost': [1.0, 2.0, -1.0, 5.0, 0.0]})

        >>> si = InterfaceToSolver()

        >>> si.add_variables(decision_variables)

        >>> si.add_objective_function(objective_function)

        >>> print(si.mip_model.var_by_name('0').obj)
        1.0

        >>> print(si.mip_model.var_by_name('5').obj)
        0.0

        """
        objective_function = objective_function.sort_values('variable_id')
        objective_function = objective_function.set_index('variable_id')
        obj = minimize(xsum(objective_function['cost'][i] * self.variables[i] for i in
                            list(objective_function.index)))
        self.mip_model.objective = obj
        self.linear_mip_model.objective = obj

    def add_constraints(self, constraints_lhs, constraints_type_and_rhs):
        """Add constraints to the mip model.

        Examples
        --------
        >>> decision_variables = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'lower_bound': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ...   'upper_bound': [5.0, 5.0, 10.0, 10.0, 5.0, 5.0],
        ...   'type': ['continuous', 'continuous', 'continuous',
        ...            'continuous', 'continuous', 'continuous']})

        >>> constraints_lhs = pd.DataFrame({
        ...   'constraint_id': [1, 1, 2, 2],
        ...   'variable_id': [0, 1, 3, 4],
        ...   'coefficient': [1.0, 0.5, 1.0, 2.0]})

        >>> constraints_type_and_rhs = pd.DataFrame({
        ...   'constraint_id': [1, 2],
        ...   'type': ['<=', '='],
        ...   'rhs': [10.0, 20.0]})

        >>> si = InterfaceToSolver()

        >>> si.add_variables(decision_variables)

        >>> si.add_constraints(constraints_lhs, constraints_type_and_rhs)

        >>> print(si.mip_model.constr_by_name('1'))
        1: +1.0 0 +0.5 1 <= 10.0

        >>> print(si.mip_model.constr_by_name('2'))
        2: +1.0 3 +2.0 4 = 20.0

        """

        constraints_lhs = constraints_lhs.groupby(['constraint_id', 'variable_id'], as_index=False).agg(
            {'coefficient': 'sum'})
        rows = constraints_lhs.groupby(['constraint_id'], as_index=False)

        # Make a dictionary so constraint rhs values can be accessed using the constraint id.
        rhs = dict(zip(constraints_type_and_rhs['constraint_id'], constraints_type_and_rhs['rhs']))
        # Make a dictionary so constraint type can be accessed using the constraint id.
        enq_type = dict(zip(constraints_type_and_rhs['constraint_id'], constraints_type_and_rhs['type']))
        var_ids = constraints_lhs['variable_id'].to_numpy()
        vars = np.asarray(
            [self.variables[k] if k in self.variables.keys() else None for k in range(0, max(var_ids) + 1)])
        coefficients = constraints_lhs['coefficient'].to_numpy()
        for row_id, row in rows.indices.items():
            # Use the variable_ids to get mip variable objects present in the constraints
            lhs_variables = vars[var_ids[row]]
            # Use the positions of the non nan values to the lhs coefficients.
            lhs = coefficients[row]
            # Multiply and the variables by their coefficients and sum to create the lhs of the constraint.
            exp = lhs_variables * lhs
            exp = exp.tolist()
            exp = xsum(exp)
            # Add based on inequality type.
            if enq_type[row_id] == '<=':
                new_constraint = exp <= rhs[row_id]
            elif enq_type[row_id] == '>=':
                new_constraint = exp >= rhs[row_id]
            elif enq_type[row_id] == '=':
                new_constraint = exp == rhs[row_id]
            else:
                raise ValueError("Constraint type not recognised should be one of '<=', '>=' or '='.")
            self.mip_model.add_constr(new_constraint, name=str(row_id))
            self.linear_mip_model.add_constr(new_constraint, name=str(row_id))

    def optimize(self):
        """Optimize the mip model.

        If an optimal solution cannot be found and the investigate_infeasibility flag is set to True then remove
        constraints until a feasible solution is found.

        Examples
        --------
        >>> decision_variables = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'lower_bound': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ...   'upper_bound': [5.0, 5.0, 10.0, 10.0, 5.0, 5.0],
        ...   'type': ['continuous', 'continuous', 'continuous',
        ...            'continuous', 'continuous', 'continuous']})

        >>> constraints_lhs = pd.DataFrame({
        ...   'constraint_id': [1, 1, 2, 2],
        ...   'variable_id': [0, 1, 3, 4],
        ...   'coefficient': [1.0, 0.5, 1.0, 2.0]})

        >>> constraints_type_and_rhs = pd.DataFrame({
        ...   'constraint_id': [1, 2],
        ...   'type': ['<=', '='],
        ...   'rhs': [10.0, 20.0]})

        >>> si = InterfaceToSolver()

        >>> si.add_variables(decision_variables)

        >>> si.add_constraints(constraints_lhs, constraints_type_and_rhs)

        >>> si.optimize()

        >>> decision_variables['value'] = si.get_optimal_values_of_decision_variables(decision_variables)

        >>> print(decision_variables)
           variable_id  lower_bound  upper_bound        type  value
        0            0          0.0          5.0  continuous    0.0
        1            1          0.0          5.0  continuous    0.0
        2            2          0.0         10.0  continuous    0.0
        3            3          0.0         10.0  continuous   10.0
        4            4          0.0          5.0  continuous    5.0
        5            5          0.0          5.0  continuous    0.0
        """
        status = self.mip_model.optimize()
        if status != OptimizationStatus.OPTIMAL:
            # Attempt find constraint causing infeasibility.
            print('Model infeasible attempting to find problem constraint.')
            con_index = find_problem_constraint(self.mip_model)
            print('Couldn\'t find an optimal solution, but removing con {} fixed INFEASIBLITY'.format(con_index))
            raise ValueError('Linear program infeasible')

    def get_optimal_values_of_decision_variables(self, variable_definitions):
        """Get the optimal values for each decision variable.

        Examples
        --------

        >>> decision_variables = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'lower_bound': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ...   'upper_bound': [5.0, 5.0, 10.0, 10.0, 5.0, 5.0],
        ...   'type': ['continuous', 'continuous', 'continuous',
        ...            'continuous', 'continuous', 'continuous']})

        >>> constraints_lhs = pd.DataFrame({
        ...   'constraint_id': [1, 1, 2, 2],
        ...   'variable_id': [0, 1, 3, 4],
        ...   'coefficient': [1.0, 0.5, 1.0, 2.0]})

        >>> constraints_type_and_rhs = pd.DataFrame({
        ...   'constraint_id': [1, 2],
        ...   'type': ['<=', '='],
        ...   'rhs': [10.0, 20.0]})

        >>> si = InterfaceToSolver()

        >>> si.add_variables(decision_variables)

        >>> si.add_constraints(constraints_lhs, constraints_type_and_rhs)

        >>> si.optimize()

        >>> decision_variables['value'] = si.get_optimal_values_of_decision_variables(decision_variables)

        >>> print(decision_variables)
           variable_id  lower_bound  upper_bound        type  value
        0            0          0.0          5.0  continuous    0.0
        1            1          0.0          5.0  continuous    0.0
        2            2          0.0         10.0  continuous    0.0
        3            3          0.0         10.0  continuous   10.0
        4            4          0.0          5.0  continuous    5.0
        5            5          0.0          5.0  continuous    0.0

        """
        values = variable_definitions['variable_id'].apply(lambda x: self.mip_model.var_by_name(str(x)).x,
                                                           self.mip_model)
        return values

    def get_optimal_values_of_decision_variables_lin(self, variable_definitions):
        values = variable_definitions['variable_id'].apply(lambda x: self.linear_mip_model.var_by_name(str(x)).x,
                                                           self.mip_model)
        return values

    def get_slack_in_constraints(self, constraints_type_and_rhs):
        """Get the slack values in each constraint.

        Examples
        --------

        >>> decision_variables = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'lower_bound': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ...   'upper_bound': [5.0, 5.0, 10.0, 10.0, 5.0, 5.0],
        ...   'type': ['continuous', 'continuous', 'continuous',
        ...            'continuous', 'continuous', 'continuous']})

        >>> constraints_lhs = pd.DataFrame({
        ...   'constraint_id': [1, 1, 2, 2],
        ...   'variable_id': [0, 1, 3, 4],
        ...   'coefficient': [1.0, 0.5, 1.0, 2.0]})

        >>> constraints_type_and_rhs = pd.DataFrame({
        ...   'constraint_id': [1, 2],
        ...   'type': ['<=', '='],
        ...   'rhs': [10.0, 20.0]})

        >>> si = InterfaceToSolver()

        >>> si.add_variables(decision_variables)

        >>> si.add_constraints(constraints_lhs, constraints_type_and_rhs)

        >>> si.optimize()

        >>> constraints_type_and_rhs['slack'] = si.get_slack_in_constraints(constraints_type_and_rhs)

        >>> print(constraints_type_and_rhs)
           constraint_id type   rhs  slack
        0              1   <=  10.0   10.0
        1              2    =  20.0    0.0

        """
        slack = constraints_type_and_rhs['constraint_id'].apply(lambda x: self.mip_model.constr_by_name(str(x)).slack,
                                                                self.mip_model)
        return slack

    def price_constraints(self, constraint_ids_to_price):
        """For each constraint_id find the marginal value of the constraint.

        This is done by incrementing the constraint by a value of 1.0 and re-optimizing the model, the marginal cost
        of the constraint is increase in the objective function value between model runs.

        Examples
        --------

        >>> decision_variables = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'lower_bound': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ...   'upper_bound': [5.0, 5.0, 10.0, 10.0, 5.0, 5.0],
        ...   'type': ['continuous', 'continuous', 'continuous',
        ...            'continuous', 'continuous', 'continuous']})

        >>> objective_function = pd.DataFrame({
        ...   'variable_id': [0, 1, 2, 3, 4, 5],
        ...   'cost': [1.0, 3.0, 10.0, 8.0, 9.0, 7.0]})

        >>> constraints_lhs = pd.DataFrame({
        ...   'constraint_id': [1, 1, 1, 1],
        ...   'variable_id': [0, 1, 3, 4],
        ...   'coefficient': [1.0, 1.0, 1.0, 1.0]})

        >>> constraints_type_and_rhs = pd.DataFrame({
        ...   'constraint_id': [1],
        ...   'type': ['='],
        ...   'rhs': [20.0]})

        >>> si = InterfaceToSolver()

        >>> si.add_variables(decision_variables)

        >>> si.add_constraints(constraints_lhs, constraints_type_and_rhs)

        >>> si.add_objective_function(objective_function)

        >>> si.optimize()

        >>> si.linear_mip_model.optimize()
        <OptimizationStatus.OPTIMAL: 0>

        >>> prices = si.price_constraints([1])

        >>> print(prices)
        {1: 8.0}

        >>> decision_variables['value'] = si.get_optimal_values_of_decision_variables(decision_variables)

        >>> print(decision_variables)
           variable_id  lower_bound  upper_bound        type  value
        0            0          0.0          5.0  continuous    5.0
        1            1          0.0          5.0  continuous    5.0
        2            2          0.0         10.0  continuous    0.0
        3            3          0.0         10.0  continuous   10.0
        4            4          0.0          5.0  continuous    0.0
        5            5          0.0          5.0  continuous    0.0

        """
        costs = {}
        for id in constraint_ids_to_price:
            costs[id] = self.linear_mip_model.constr_by_name(str(id)).pi
        return costs

    def update_rhs(self, constraint_id, violation_degree):
        constraint = self.linear_mip_model.constr_by_name(str(constraint_id))
        constraint.rhs += violation_degree

    def update_variable_bounds(self, new_bounds):
        for variable_id, lb, ub in zip(new_bounds['variable_id'], new_bounds['lower_bound'], new_bounds['upper_bound']):
            self.mip_model.var_by_name(str(variable_id)).lb = lb
            self.mip_model.var_by_name(str(variable_id)).ub = ub

    def disable_variables(self, variables):
        for var_id in variables['variable_id']:
            var = self.linear_mip_model.var_by_name(str(var_id))
            var.lb = 0.0
            var.ub = 0.0


def find_problem_constraint(base_prob):
    cons = []
    test_prob = base_prob.copy()
    for con in [con.name for con in base_prob.constrs]:
        [test_prob.remove(c) for c in test_prob.constrs if c.name == con]
        status = test_prob.optimize()
        cons.append(con)
        if status == OptimizationStatus.OPTIMAL:
            return cons
    return []


def create_lhs(constraints, decision_variables, join_columns):
    """Combine constraints with general definitions of lhs with variables to give an explicit lhs definition.

    Both constraints and decision_variables can have a coefficient, the coefficient use in the actual lhs will
    be the product of the two coefficients.

    Examples
    --------

    >>> decision_variables = pd.DataFrame({
    ...   'variable_id': [0, 1, 2, 3, 4, 5],
    ...   'region': ['NSW', 'NSW', 'VIC',
    ...              'VIC', 'VIC', 'VIC'],
    ...   'service': ['energy', 'energy','energy',
    ...               'energy','energy','energy',],
    ...   'coefficient': [0.9, 0.8, 1.0, 0.95, 1.1, 1.01]})

    >>> constraints = pd.DataFrame({
    ...   'constraint_id': [1, 2],
    ...   'region': ['NSW', 'VIC'],
    ...   'service': ['energy', 'energy'],
    ...   'coefficient': [1.0, 1.0]})

    >>> lhs = create_lhs(decision_variables, constraints, ['region', 'service'])

    >>> print(lhs)
       constraint_id  variable_id  coefficient
    0              1            0         0.90
    1              1            1         0.80
    2              2            2         1.00
    3              2            3         0.95
    4              2            4         1.10
    5              2            5         1.01

    Parameters
    ----------
    constraints : pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        join_columns   one or more columns defining the types of variables that should
                       be on the lhs (as `str`)
        coefficient    the constraint level contribution to the lhs coefficient (as `np.float64`)
        =============  ===============================================================

    decision_variables : pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        variable_id    the unique identifier of the variable (as `np.int64`)
        join_columns   one or more columns defining the types of variables that should
                       be on the lhs (as `str`)
        coefficient    the variable level contribution to the lhs coefficient (as `np.float64`)
        =============  ===============================================================

    Returns
    -------
    lhs : pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        variable_id    the unique identifier of the variable (as `np.int64`)
        coefficient    the constraint level contribution to the lhs coefficient (as `np.float64`)
        =============  ===============================================================
    """
    constraints = pd.merge(constraints, decision_variables, 'inner', on=join_columns)
    constraints['coefficient'] = constraints['coefficient_x'] * constraints['coefficient_y']
    lhs = constraints.loc[:, ['constraint_id', 'variable_id', 'coefficient']]
    return lhs


def create_mapping_of_generic_constraint_sets_to_constraint_ids(constraints, market_constraints):
    """Combine generic constraints and fcas market constraints to get the full set of generic constraints.

    Returns non if there are no generic of fcas market constraints.

    Examples
    --------

    >>> constraints = {
    ...   'generic': pd.DataFrame({
    ...       'constraint_id': [0, 1],
    ...       'set': ['A', 'B']})
    ...   }

    >>> market_constraints = {
    ...   'fcas': pd.DataFrame({
    ...       'constraint_id': [2, 3],
    ...       'set': ['C', 'D']})
    ...   }

    >>> generic_constraints = create_mapping_of_generic_constraint_sets_to_constraint_ids(
    ... constraints, market_constraints)

    >>> print(generic_constraints)
       constraint_id set
    0              0   A
    1              1   B
    0              2   C
    1              3   D


    Parameters
    ----------
    constraints : dict{str : pd.DataFrame}

        The pd.DataFrame stored under the key 'generic', if it exists, should have the structure.

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        set            the constraint set that the id refers to (as `str`)
        =============  ===============================================================

    market_constraints : dict{str : pd.DataFrame}

        The pd.DataFrame stored under the key 'fcas', if it exists, should have the structure.

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        set            the constraint set that the id refers to (as `str`)
        =============  ===============================================================

    Returns
    -------
    pd.DataFrame or None

        If pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        set            the constraint set that the id refers to (as `str`)
        =============  ===============================================================

    """
    generic_constraints = []
    if 'generic' in constraints:
        generic_constraints.append(constraints['generic'].loc[:, ['constraint_id', 'set']])
    if 'fcas' in market_constraints:
        generic_constraints.append(market_constraints['fcas'].loc[:, ['constraint_id', 'set']])
    if len(generic_constraints) > 0:
        return pd.concat(generic_constraints)
    else:
        return None


def create_unit_level_generic_constraint_lhs(generic_constraint_units, generic_constraint_ids,
                                             unit_bids_to_constraint_map):
    """Find the lhs variables from units for generic constraints.

    Examples
    --------

    >>> generic_constraint_units = pd.DataFrame({
    ...   'set': ['A', 'A'],
    ...   'unit': ['X', 'Y'],
    ...   'service': ['energy', 'energy'],
    ...   'coefficient': [0.9, 0.8]})

    >>> generic_constraint_ids = pd.DataFrame({
    ...   'constraint_id': [1, 2],
    ...   'set': ['A', 'B']})

    >>> unit_bids_to_constraint_map = pd.DataFrame({
    ...   'variable_id': [0, 1],
    ...   'unit': ['X', 'Y'],
    ...   'service': ['energy', 'energy']})

    >>> lhs = create_unit_level_generic_constraint_lhs(generic_constraint_units, generic_constraint_ids,
    ...   unit_bids_to_constraint_map)

    >>> print(lhs)
       constraint_id  variable_id  coefficient
    0              1            0          0.9
    1              1            1          0.8

    Parameters
    ----------
    generic_constraint_units : pd.DataFrame

        =============  ==============================================================
        Columns:       Description:
        set            the unique identifier of the constraint set to map the
                       lhs coefficients to (as `str`)
        unit           the unit whose variables will be mapped to the lhs (as `str`)
        service        the service whose variables will be mapped to the lhs (as `str`)
        coefficient    the lhs coefficient (as `np.float64`)
        =============  ==============================================================

    generic_constraint_ids : pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        set            the constraint set that the id refers to (as `str`)
        =============  ===============================================================

    unit_bids_to_constraint_map : pd.DataFrame

        =============  =============================================================================
        Columns:       Description:
        variable_id    the id of the variable (as `np.int64`)
        unit           the unit level constraints the variable should map to (as `str`)
        service        the service type of the constraints the variables should map to (as `str`)
        =============  =============================================================================

    Returns
    -------
    lhs : pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        variable_id    the unique identifier of the variable (as `np.int64`)
        coefficient    the constraint level contribution to the lhs coefficient (as `np.float64`)
        =============  ===============================================================
    """
    unit_lhs = pd.merge(generic_constraint_units,
                        unit_bids_to_constraint_map.loc[:, ['unit', 'service', 'variable_id']],
                        on=['unit', 'service'])
    unit_lhs = pd.merge(unit_lhs, generic_constraint_ids.loc[:, ['constraint_id', 'set']], on='set')
    return unit_lhs.loc[:, ['constraint_id', 'variable_id', 'coefficient']]


def create_region_level_generic_constraint_lhs(generic_constraint_regions, generic_constraint_ids,
                                               regional_bids_to_constraint_map):
    """Find the lhs variables from regions for generic constraints.

    Examples
    --------

    >>> generic_constraint_regions = pd.DataFrame({
    ...   'set': ['A'],
    ...   'region': ['X'],
    ...   'service': ['energy'],
    ...   'coefficient': [0.9]})

    >>> generic_constraint_ids = pd.DataFrame({
    ...   'constraint_id': [1, 2],
    ...   'set': ['A', 'B']})

    >>> regional_bids_to_constraint_map = pd.DataFrame({
    ...   'variable_id': [0, 1],
    ...   'region': ['X', 'X'],
    ...   'service': ['energy', 'energy']})

    >>> lhs = create_region_level_generic_constraint_lhs(generic_constraint_regions, generic_constraint_ids,
    ...   regional_bids_to_constraint_map)

    >>> print(lhs)
       constraint_id  variable_id  coefficient
    0              1            0          0.9
    1              1            1          0.9

    Parameters
    ----------
    generic_constraint_regions : pd.DataFrame

        =============  ==============================================================
        Columns:       Description:
        set            the unique identifier of the constraint set to map the
                       lhs coefficients to (as `str`)
        region         the region whose variables will be mapped to the lhs (as `str`)
        service        the service whose variables will be mapped to the lhs (as `str`)
        coefficient    the lhs coefficient (as `np.float64`)
        =============  ==============================================================

    generic_constraint_ids : pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        set            the constraint set that the id refers to (as `str`)
        =============  ===============================================================

    regional_bids_to_constraint_map : pd.DataFrame

        =============  =============================================================================
        Columns:       Description:
        variable_id    the id of the variable (as `np.int64`)
        region         the region level constraints the variable should map to (as `str`)
        service        the service type of the constraints the variables should map to (as `str`)
        =============  =============================================================================

    Returns
    -------
    lhs : pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        constraint_id  the unique identifier of the constraint (as `np.int64`)
        variable_id    the unique identifier of the variable (as `np.int64`)
        coefficient    the constraint level contribution to the lhs coefficient (as `np.float64`)
        =============  ===============================================================
    """
    region_lhs = pd.merge(generic_constraint_regions,
                          regional_bids_to_constraint_map.loc[:, ['region', 'service', 'variable_id']],
                          on=['region', 'service'])
    region_lhs = pd.merge(region_lhs, generic_constraint_ids.loc[:, ['constraint_id', 'set']], on='set')
    return region_lhs.loc[:, ['constraint_id', 'variable_id', 'coefficient']]


def create_interconnector_generic_constraint_lhs(generic_constraint_interconnectors, generic_constraint_ids,
                                                 interconnector_variables):
    """Find the lhs variables from interconnectors for generic constraints.

    Examples
    --------

    >>> generic_constraint_interconnectors = pd.DataFrame({
    ...   'set': ['A'],
    ...   'interconnector': ['X'],
    ...   'coefficient': [0.9]})

    >>> generic_constraint_ids = pd.DataFrame({
    ...   'constraint_id': [1, 2],
    ...   'set': ['A', 'B']})

    >>> interconnector_variables = pd.DataFrame({
    ...   'variable_id': [0, 1],
    ...   'interconnector': ['X', 'X'],
    ...   'generic_constraint_factor': [1, 1]})

    >>> lhs = create_interconnector_generic_constraint_lhs(generic_constraint_interconnectors, generic_constraint_ids,
    ...   interconnector_variables)

    >>> print(lhs)
       constraint_id  variable_id  coefficient
    0              1            0          0.9
    1              1            1          0.9
    """
    interconnector_lhs = pd.merge(generic_constraint_interconnectors,
                                  interconnector_variables.loc[:, ['interconnector', 'variable_id',
                                                                   'generic_constraint_factor']],
                                  on=['interconnector'])
    interconnector_lhs = pd.merge(interconnector_lhs, generic_constraint_ids.loc[:, ['constraint_id', 'set']], on='set')
    interconnector_lhs['coefficient'] = interconnector_lhs['coefficient'] * interconnector_lhs[
        'generic_constraint_factor']
    return interconnector_lhs.loc[:, ['constraint_id', 'variable_id', 'coefficient']]



# Notice:
# - This script downloads large volumes of historical market data from AEMO's nemweb
#   portal. The boolean on line 20 can be changed to prevent this happening repeatedly
#   once the data has been downloaded.

import sqlite3
import pandas as pd
import random
from datetime import datetime, timedelta

from nempy import markets
from nempy.historical_inputs import loaders, mms_db, \
    xml_cache, units, demand, interconnectors, \
    constraints

# The size of historical data files for a full year of 5 min dispatch
# is very large, approximately 800 GB, for this reason the data is
# stored on an external SSD.
con = sqlite3.connect('historical_mms.db')
mms_db_manager = mms_db.DBManager(connection=con)
xml_cache_manager = xml_cache.XMLCacheManager('nemde_cache')

# The second time this example is run on a machine this flag can
# be set to false to save downloading the data again.
download_inputs = True

if download_inputs:
    # This requires approximately 5 GB of storage.
    mms_db_manager.populate(start_year=2019, start_month=1,
                            end_year=2019, end_month=12)

    # This requires approximately 3.5 GB of storage.
    xml_cache_manager.populate(start_year=2019, start_month=1, start_day=1,
                               end_year=2020, end_month=1, end_day=1)

raw_inputs_loader = loaders.RawInputsLoader(
    nemde_xml_cache_manager=xml_cache_manager,
    market_management_system_database=mms_db_manager)


# Define a function for creating a list of randomly selected dispatch
# intervals
def get_test_intervals(number):
    start_time = datetime(year=2019, month=1, day=1, hour=0, minute=0)
    end_time = datetime(year=2019, month=12, day=31, hour=0, minute=0)
    difference = end_time - start_time
    difference_in_5_min_intervals = difference.days * 12 * 24
    random.seed(1)
    intervals = random.sample(range(1, difference_in_5_min_intervals), number)
    times = [start_time + timedelta(minutes=5 * i) for i in intervals]
    times_formatted = [t.isoformat().replace('T', ' ').replace('-', '/') for t in times]
    return times_formatted


# List for saving outputs to.
outputs = []

# Create and dispatch the spot market for each dispatch interval.
for interval in get_test_intervals(number=1000):
    raw_inputs_loader.set_interval(interval)
    unit_inputs = units.UnitData(raw_inputs_loader)
    interconnector_inputs = interconnectors.InterconnectorData(raw_inputs_loader)
    constraint_inputs = constraints.ConstraintData(raw_inputs_loader)
    demand_inputs = demand.DemandData(raw_inputs_loader)

    unit_info = unit_inputs.get_unit_info()
    market = markets.SpotMarket(market_regions=['QLD1', 'NSW1', 'VIC1',
                                                'SA1', 'TAS1'],
                                unit_info=unit_info)

    # By default the CBC open source solver is used, but GUROBI is
    # also supported
    market.solver_name = 'CBC'  # or could be 'GUROBI'

    # Set bids
    volume_bids, price_bids = unit_inputs.get_processed_bids()
    volume_bids = volume_bids[volume_bids['service'] == 'energy']
    price_bids = price_bids[price_bids['service'] == 'energy']
    market.set_unit_volume_bids(volume_bids)
    market.set_unit_price_bids(price_bids)

    # Set bid in capacity limits
    unit_bid_limit = unit_inputs.get_unit_bid_availability()
    market.set_unit_bid_capacity_constraints(unit_bid_limit)
    cost = constraint_inputs.get_constraint_violation_prices()['unit_capacity']
    market.make_constraints_elastic('unit_bid_capacity', violation_cost=cost)

    # Set limits provided by the unconstrained intermittent generation
    # forecasts. Primarily for wind and solar.
    unit_uigf_limit = unit_inputs.get_unit_uigf_limits()
    market.set_unconstrained_intermitent_generation_forecast_constraint(
        unit_uigf_limit)
    cost = constraint_inputs.get_constraint_violation_prices()['uigf']
    market.make_constraints_elastic('uigf_capacity', violation_cost=cost)

    # Set unit ramp rates.
    ramp_rates = unit_inputs.get_ramp_rates_used_for_energy_dispatch()
    market.set_unit_ramp_up_constraints(
        ramp_rates.loc[:, ['unit', 'initial_output', 'ramp_up_rate']])
    market.set_unit_ramp_down_constraints(
        ramp_rates.loc[:, ['unit', 'initial_output', 'ramp_down_rate']])
    cost = constraint_inputs.get_constraint_violation_prices()['ramp_rate']
    market.make_constraints_elastic('ramp_up', violation_cost=cost)
    market.make_constraints_elastic('ramp_down', violation_cost=cost)

    # Set interconnector definitions, limits and loss models.
    interconnectors_definitions = \
        interconnector_inputs.get_interconnector_definitions()
    loss_functions, interpolation_break_points = \
        interconnector_inputs.get_interconnector_loss_model()
    market.set_interconnectors(interconnectors_definitions)
    market.set_interconnector_losses(loss_functions,
                                     interpolation_break_points)

    # Set the operational demand to be met by dispatch.
    regional_demand = demand_inputs.get_operational_demand()
    market.set_demand_constraints(regional_demand)
    market.dispatch()

    # Save prices from this interval
    prices = market.get_energy_prices()
    prices['time'] = interval

    # Getting historical prices for comparison. Note, ROP price, which is
    # the regional reference node price before the application of any
    # price scaling by AEMO, is used for comparison.
    historical_prices = mms_db_manager.DISPATCHPRICE.get_data(interval)

    prices = pd.merge(prices, historical_prices,
                      left_on=['time', 'region'],
                      right_on=['SETTLEMENTDATE', 'REGIONID'])

    outputs.append(
        prices.loc[:, ['time', 'region', 'price',
                       'SETTLEMENTDATE', 'REGIONID', 'ROP']])

con.close()
outputs = pd.concat(outputs)
outputs = outputs.sort_values('ROP')
outputs = outputs.reset_index(drop=True)
outputs.to_csv('energy_price_results_2019_1000_intervals_without_FCAS_or_generic_constraints.csv')




import pickle

from nempy.historical_inputs import xml_cache

import os
cwd = os.getcwd()

print(cwd)

historical_inputs = xml_cache.XMLCacheManager('test_files/historical_xml_files')

interval_with_violations = \
    historical_inputs.find_intervals_with_violations(limit=1000000,
                                                     start_year=2019, start_month=2,
                                                     end_year=2019, end_month=2)

with open('interval_with_violations.pickle', 'wb') as f:
    pickle.dump(interval_with_violations, f, pickle.HIGHEST_PROTOCOL)



import pandas as pd
import numpy as np
from nempy.help_functions import helper_functions as hf


def bids(variable_ids, price_bids, unit_info):
    """Create the cost coefficients of energy in bids in the objective function.

    This function defines the cost associated with each decision variable that represents a unit's energy bid. Costs are
    are with reference to the regional node.
    """
    # If no service column is provided assume bids are for energy.
    if 'service' not in price_bids.columns:
        price_bids['service'] = 'energy'

    # Get the list of columns that are bid bands.
    bid_bands = [col for col in price_bids.columns if col not in ['unit', 'service']]
    price_bids = hf.stack_columns(price_bids, cols_to_keep=['unit', 'service'], cols_to_stack=bid_bands,
                                  type_name='capacity_band', value_name='cost')
    # Match bid cost with existing variable ids
    objective_function = pd.merge(variable_ids, price_bids, how='inner', on=['unit', 'service', 'capacity_band'])
    objective_function = pd.merge(objective_function, unit_info.loc[:, ['unit', 'dispatch_type']], how='inner',
                                  on=['unit'])
    objective_function['cost'] = np.where((objective_function['dispatch_type'] == 'load') &
                                          (objective_function['service'] == 'energy'),
                                          -1.0 * objective_function['cost'], objective_function['cost'])
    return objective_function


def scale_by_loss_factors(objective_function, unit_info):
    """
    Scale the bid cost by dividing by the loss factor.

    Parameters
    ----------
    objective_function : pd.DataFrame
        Cost by variable id, also including unit and capacity band so loss factors can be applied if provided.

        =============  ===============================================================
        Columns:       Description:
        unit           unique identifier of a dispatch unit (as `str`)
        capacity_band  the bid band of the variable (as `str`)
        variable_id    the id of the variable (as `int`)
        =============  ===============================================================

    unit_info : pd.DataFrame
        The loss factor to scale bids by.

        =============  ===============================================================
        Columns:       Description:
        unit           unique identifier of a dispatch unit (as `str`)
        loss_factor    the id of the variable (as `int`)
        =============  ===============================================================

    Returns
    -------
    pd.DataFrame

        =============  ===============================================================
        Columns:       Description:
        unit           unique identifier of a dispatch unit (as `str`)
        capacity_band  the bid band of the variable (as `str`)
        variable_id    the id of the variable (as `int`)
        cost           the bid cost of the variable (as `float`)
        =============  ===============================================================
    """

    # Match units with their loss factors.
    objective_function = pd.merge(objective_function, unit_info, how='inner', on='unit')
    # Refer bids cost to regional reference node, if a loss factor  was provided.
    objective_function['cost'] = np.where(objective_function['service'] == 'energy',
                                          objective_function['cost'] / objective_function['loss_factor'],
                                          objective_function['cost'])
    return objective_function



import pytest
import pandas as pd
from nempy.historical_inputs import interconnectors


def test_create_loss_function():
    # Interconnector flow
    flow = 600.0

    # Arbitrary demand inputs
    nsw_demand = 7000.0
    qld_demand = 5000.0
    demand = pd.DataFrame({
        'region': ['NSW1', 'QLD1'],
        'loss_function_demand': [nsw_demand, qld_demand]
    })

    # Loss model details from 2020 Jan NEM web files.
    demand_coefficients = pd.DataFrame({
        'interconnector': ['NSW1-QLD1', 'NSW1-QLD1'],
        'region': ['NSW1', 'QLD1'],
        'demand_coefficient': [-0.00000035146, 0.000010044]
    })

    # Loss model details from 2020 Jan NEM web files.
    interconnector_coefficients = pd.DataFrame({
        'interconnector': ['NSW1-QLD1'],
        'loss_constant': [0.9529],
        'flow_coefficient': [0.00019617],
        'from_region_loss_share': [0.5]
    })

    loss_function = interconnectors.create_loss_functions(interconnector_coefficients,
                                                          demand_coefficients, demand)

    output_losses = loss_function['loss_function'].loc[0](flow)

    expected_losses = (-0.0471 - 3.5146E-07 * nsw_demand + 1.0044E-05 * qld_demand) * flow + 9.8083E-05 * flow ** 2

    assert(pytest.approx(expected_losses, 0.0001) == output_losses)


def test_create_loss_function_vic_nsw():
    # Interconnector flow
    flow = 600.0

    # Arbitrary demand inputs
    vic_demand = 6000.0
    nsw_demand = 7000.0
    sa_demand = 3000.0
    demand = pd.DataFrame({
        'region': ['NSW1', 'VIC1', 'SA1'],
        'loss_function_demand': [nsw_demand, vic_demand, sa_demand]
    })

    # Loss model details from 2020 Jan NEM web files.
    demand_coefficients = pd.DataFrame({
        'interconnector': ['VIC1-NSW1', 'VIC1-NSW1', 'VIC1-NSW1'],
        'region': ['NSW1', 'VIC1', 'SA1'],
        'demand_coefficient': [0.000021734, -0.000031523, -0.000065967]
    })

    # Loss model details from 2020 Jan NEM web files.
    interconnector_coefficients = pd.DataFrame({
        'interconnector': ['VIC1-NSW1'],
        'loss_constant': [1.0657],
        'flow_coefficient': [0.00017027],
        'from_region_loss_share': [0.5]
    })

    loss_function = interconnectors.create_loss_functions(interconnector_coefficients,
                                                          demand_coefficients, demand)

    output_losses = loss_function['loss_function'].loc[0](flow)

    expected_losses = (0.0657 - 3.1523E-05 * vic_demand + 2.1734E-05 * nsw_demand
                       - 6.5967E-05 * sa_demand) * flow + 8.5133E-05 * flow ** 2

    assert (pytest.approx(expected_losses, 0.0001) == output_losses)


def test_create_loss_function_bass_link():
    # Interconnector flow
    flow = -433.0

    # Arbitrary demand inputs
    nsw_demand = 6000.0
    qld_demand = 7000.0
    demand = pd.DataFrame({
        'region': ['VIC1', 'QLD1'],
        'loss_function_demand': [nsw_demand, qld_demand]
    })

    # Loss model details from 2020 Jan NEM web files.
    demand_coefficients = pd.DataFrame({
        'interconnector': ['BL'],
        'region': ['VIC1'],
        'demand_coefficient': [0.0]
    })

    # Loss model details from 2020 Jan NEM web files.
    interconnector_coefficients = pd.DataFrame({
        'interconnector': ['BL'],
        'loss_constant': [0.99608],
        'flow_coefficient': [0.00020786],
        'from_region_loss_share': [0.5]
    })

    loss_function = interconnectors.create_loss_functions(interconnector_coefficients,
                                                          demand_coefficients, demand)

    output_losses = loss_function['loss_function'].loc[0](flow)

    expected_losses = (-3.92E-3) * flow + (1.0393E-4) * flow ** 2

    assert(pytest.approx(expected_losses, 0.0001) == output_losses)


import pandas as pd
from pandas._testing import assert_frame_equal
from nempy.spot_markert_backend import unit_constraints


def test_create_constraints():
    unit_limit = pd.DataFrame({
        'unit': ['A', 'B'],
        'upper': [16.0, 23.0]
    })
    next_constraint_id = 4
    rhs_col = 'upper'
    direction = '<='
    output_rhs, output_variable_map = unit_constraints.create_constraints(unit_limit, next_constraint_id, rhs_col,
                                                                          direction)
    expected_rhs = pd.DataFrame({
        'unit': ['A', 'B'],
        'service': ['energy', 'energy'],
        'constraint_id': [4, 5],
        'type': ['<=', '<='],
        'rhs': [16.0, 23.0]
    })
    expected_variable_map = pd.DataFrame({
        'constraint_id': [4, 5],
        'unit': ['A', 'B'],
        'service': ['energy', 'energy'],
        'coefficient': [1.0, 1.0]
    })
    assert_frame_equal(output_rhs, expected_rhs)
    assert_frame_equal(output_variable_map, expected_variable_map)


def test_one_unit_create_constraints():
    unit_limit = pd.DataFrame({
        'unit': ['A'],
        'upper': [16.0]
    })
    next_constraint_id = 4
    rhs_col = 'upper'
    direction = '<='
    output_rhs, output_variable_map = unit_constraints.create_constraints(unit_limit, next_constraint_id, rhs_col,
                                                                          direction)
    expected_rhs = pd.DataFrame({
        'unit': ['A'],
        'service': ['energy'],
        'constraint_id': [4],
        'type': ['<='],
        'rhs': [16.0],

    })
    expected_variable_map = pd.DataFrame({
        'constraint_id': [4],
        'unit': ['A'],
        'service': ['energy'],
        'coefficient': [1.0],
    })
    assert_frame_equal(output_rhs, expected_rhs)
    assert_frame_equal(output_variable_map, expected_variable_map)


def test_ramp_down():
    unit_limit = pd.DataFrame({
        'unit': ['A', 'B'],
        'initial_output': [16.0, 23.0],
        'ramp_down_rate': [12.0, 36.0]
    })
    next_constraint_id = 4
    dispatch_interval = 5
    output_rhs, output_variable_map = unit_constraints.ramp_down(unit_limit, next_constraint_id, dispatch_interval)
    expected_rhs = pd.DataFrame({
        'unit': ['A', 'B'],
        'service': ['energy', 'energy'],
        'constraint_id': [4, 5],
        'type': ['>=', '>='],
        'rhs': [15.0, 20.0]
    })
    expected_variable_map = pd.DataFrame({
        'constraint_id': [4, 5],
        'unit': ['A', 'B'],
        'service': ['energy', 'energy'],
        'coefficient': [1.0, 1.0],
    })
    assert_frame_equal(output_rhs.reset_index(drop=True), expected_rhs)
    assert_frame_equal(output_variable_map.reset_index(drop=True), expected_variable_map)


def test_ramp_up():
    unit_limit = pd.DataFrame({
        'unit': ['A', 'B'],
        'initial_output': [16, 23],
        'ramp_up_rate': [12, 36]
    })
    next_constraint_id = 4
    dispatch_interval = 5
    output_rhs, output_variable_map = unit_constraints.ramp_up(unit_limit, next_constraint_id, dispatch_interval)
    expected_rhs = pd.DataFrame({
        'unit': ['A', 'B'],
        'service': ['energy', 'energy'],
        'constraint_id': [4, 5],
        'type': ['<=', '<='],
        'rhs': [17.0, 26.0],
    })
    expected_variable_map = pd.DataFrame({
        'constraint_id': [4, 5],
        'unit': ['A', 'B'],
        'service': ['energy', 'energy'],
        'coefficient': [1.0, 1.0]
    })
    assert_frame_equal(output_rhs.reset_index(drop=True), expected_rhs)
    assert_frame_equal(output_variable_map.reset_index(drop=True), expected_variable_map)


def test_capacity():
    unit_limit = pd.DataFrame({
        'unit': ['A', 'B'],
        'capacity': [16.0, 23.0]
    })
    next_constraint_id = 4
    output_rhs, output_variable_map = unit_constraints.capacity(unit_limit, next_constraint_id)
    expected_rhs = pd.DataFrame({
        'unit': ['A', 'B'],
        'service': ['energy', 'energy'],
        'constraint_id': [4, 5],
        'type': ['<=', '<='],
        'rhs': [16.0, 23.0],
    })
    expected_variable_map = pd.DataFrame({
        'constraint_id': [4, 5],
        'unit': ['A', 'B'],
        'service': ['energy', 'energy'],
        'coefficient': [1.0, 1.0],
    })
    assert_frame_equal(output_rhs.reset_index(drop=True), expected_rhs)
    assert_frame_equal(output_variable_map.reset_index(drop=True), expected_variable_map)



import pandas as pd
import numpy as np

from nempy.historical_inputs import demand, aemo_to_nempy_name_mapping


def _test_setup():
    import sqlite3
    from nempy.historical_inputs import mms_db
    from nempy.historical_inputs import xml_cache
    from nempy.historical_inputs import loaders
    con = sqlite3.connect('market_management_system.db')
    mms_db_manager = mms_db.DBManager(connection=con)
    xml_cache_manager = xml_cache.XMLCacheManager('test_nemde_cache')
    inputs_loader = loaders.RawInputsLoader(xml_cache_manager, mms_db_manager)
    inputs_loader.set_interval('2019/01/10 12:05:00')
    return inputs_loader


class InterconnectorData:
    """Loads interconnector related raw inputs and preprocess them for compatibility with :class:`nempy.markets.SpotMarket`

    Examples
    --------

    This example shows the setup used for the examples in the class methods. This setup is used to create a
    RawInputsLoader by calling the function _test_setup.

    >>> import sqlite3
    >>> from nempy.historical_inputs import mms_db
    >>> from nempy.historical_inputs import xml_cache
    >>> from nempy.historical_inputs import loaders

    The InterconnectorData class requries a RawInputsLoader instance.

    >>> con = sqlite3.connect('market_management_system.db')
    >>> mms_db_manager = mms_db.DBManager(connection=con)
    >>> xml_cache_manager = xml_cache.XMLCacheManager('test_nemde_cache')
    >>> inputs_loader = loaders.RawInputsLoader(xml_cache_manager, mms_db_manager)
    >>> inputs_loader.set_interval('2019/01/10 12:05:00')

    Create a InterconnectorData instance.

    >>> interconnector_data = InterconnectorData(inputs_loader)

    >>> interconnector_data.get_interconnector_definitions()
      interconnector from_region to_region     min     max       link  from_region_loss_factor  to_region_loss_factor  generic_constraint_factor
    0           V-SA        VIC1       SA1  -850.0   950.0       V-SA                   1.0000                 1.0000                          1
    1      N-Q-MNSP1        NSW1      QLD1  -264.0   264.0  N-Q-MNSP1                   1.0000                 1.0000                          1
    2      NSW1-QLD1        NSW1      QLD1 -1659.0  1229.0  NSW1-QLD1                   1.0000                 1.0000                          1
    3      V-S-MNSP1        VIC1       SA1  -270.0   270.0  V-S-MNSP1                   1.0000                 1.0000                          1
    5      VIC1-NSW1        VIC1      NSW1 -2299.0  2399.0  VIC1-NSW1                   1.0000                 1.0000                          1
    0      T-V-MNSP1        TAS1      VIC1     0.0   478.0    BLNKTAS                   1.0000                 0.9839                          1
    1      T-V-MNSP1        VIC1      TAS1     0.0   478.0    BLNKVIC                   0.9839                 1.0000                         -1

    Parameters
    ----------
    inputs_manager : historical_spot_market_inputs.DBManager
    """
    def __init__(self, raw_input_loader):
        self.raw_input_loader = raw_input_loader

        self.INTERCONNECTORCONSTRAINT = self.raw_input_loader.get_interconnector_constraint_parameters()

        # The from region loss share for Basslink is not properly defined in the AEMO data sources, for nempy to best
        # replicate NEMDE outcomes the from region loss share is set to one.
        self.INTERCONNECTORCONSTRAINT['FROMREGIONLOSSSHARE'] = \
            np.where(self.INTERCONNECTORCONSTRAINT['INTERCONNECTORID'] == 'T-V-MNSP1', 1.0,
                     self.INTERCONNECTORCONSTRAINT['FROMREGIONLOSSSHARE'])

        self.INTERCONNECTOR = self.raw_input_loader.get_interconnector_definitions()

        self.interconnectors = _format_interconnector_definitions(self.INTERCONNECTOR, self.INTERCONNECTORCONSTRAINT)

    def get_interconnector_loss_model(self):
        """Returns inputs in the format needed to set interconnector losses in the SpotMarket class.

        Examples
        --------

        >>> inputs_loader = _test_setup()

        >>> interconnector_data = InterconnectorData(inputs_loader)
        
        >>> loss_function, interpolation_break_points = \
             interconnector_data.get_interconnector_loss_model()

        >>> print(loss_function)
          interconnector       link                                      loss_function  from_region_loss_share
        0           V-SA       V-SA  <function InterconnectorData.get_interconnecto...                    0.78
        1      N-Q-MNSP1  N-Q-MNSP1  <function InterconnectorData.get_interconnecto...                    0.66
        2      NSW1-QLD1  NSW1-QLD1  <function InterconnectorData.get_interconnecto...                    0.68
        3      V-S-MNSP1  V-S-MNSP1  <function InterconnectorData.get_interconnecto...                    0.67
        4      VIC1-NSW1  VIC1-NSW1  <function InterconnectorData.get_interconnecto...                    0.32
        5      T-V-MNSP1    BLNKTAS  <function InterconnectorData.get_interconnecto...                    1.00
        6      T-V-MNSP1    BLNKVIC  <function InterconnectorData.get_interconnecto...                    1.00

        >>> print(interpolation_break_points)
            interconnector     link  loss_segment  break_point
        0             V-SA     V-SA             1       -851.0
        1             V-SA     V-SA             2       -835.0
        2             V-SA     V-SA             3       -820.0
        3             V-SA     V-SA             4       -805.0
        4             V-SA     V-SA             5       -790.0
        ..             ...      ...           ...          ...
        599      T-V-MNSP1  BLNKVIC           -80       -546.0
        600      T-V-MNSP1  BLNKVIC           -81       -559.0
        601      T-V-MNSP1  BLNKVIC           -82       -571.0
        602      T-V-MNSP1  BLNKVIC           -83       -583.0
        603      T-V-MNSP1  BLNKVIC           -84       -595.0
        <BLANKLINE>
        [604 rows x 4 columns]

        Multiple Returns
        ----------------

        loss_functions : pd.DataFrame

            ======================  ==================================
            Columns:                Description:
            interconnector          unique identifier of a interconnector, \n
                                    (as `str`)
            from_region_loss_share  The fraction of loss occuring in \n
                                    the from region, 0.0 to 1.0, \n
                                    (as `np.float64`)
            loss_function           A function that takes a flow, \n
                                    in MW as a float and returns the \n
                                    losses in MW, (as `callable`)
            ======================  ==================================

        interpolation_break_points : pd.DataFrame

            ==============  ==========================================
            Columns:        Description:
            interconnector  unique identifier of a interconnector, \n
                            (as `str`)
            loss_segment    unique identifier of a loss segment on \n
                            an interconnector basis, (as `np.float64`)
            break_point     points between which the loss function \n
                            will be linearly interpolated, in MW \n
                            (as `np.float64`)
            ==============  ==========================================

        """

        DISPATCHREGIONSUM = self.raw_input_loader.get_regional_loads()
        LOSSFACTORMODEL = self.raw_input_loader.get_interconnector_loss_parameters()
        LOSSMODEL = self.raw_input_loader.get_interconnector_loss_segments()

        regional_demand = demand._format_regional_demand(DISPATCHREGIONSUM)

        interconnector_loss_coefficients = \
            self.INTERCONNECTORCONSTRAINT.loc[:, ['INTERCONNECTORID', 'LOSSCONSTANT', 'LOSSFLOWCOEFFICIENT',
                                                  'FROMREGIONLOSSSHARE']]
        interconnector_loss_coefficients = aemo_to_nempy_name_mapping.map_aemo_column_names_to_nempy_names(
            interconnector_loss_coefficients)

        interconnector_demand_coefficients = LOSSFACTORMODEL.loc[:, ['INTERCONNECTORID', 'REGIONID',
                                                                     'DEMANDCOEFFICIENT']]
        interconnector_demand_coefficients = aemo_to_nempy_name_mapping.map_aemo_column_names_to_nempy_names(
            interconnector_demand_coefficients)


        interpolation_break_points = LOSSMODEL.loc[:, ['INTERCONNECTORID', 'LOSSSEGMENT', 'MWBREAKPOINT']]
        interpolation_break_points = aemo_to_nempy_name_mapping.map_aemo_column_names_to_nempy_names(
            interpolation_break_points)
        interpolation_break_points['loss_segment'] = interpolation_break_points['loss_segment'].apply(np.int64)


        loss_functions = create_loss_functions(interconnector_loss_coefficients,
                                               interconnector_demand_coefficients,
                                                regional_demand.loc[:, ['region', 'loss_function_demand']])

        interconnectors = self.get_interconnector_definitions()

        interpolation_break_points = pd.merge(interconnectors.loc[:, ['interconnector', 'link',
                                                                      'generic_constraint_factor']],
                                              interpolation_break_points, on='interconnector')

        interpolation_break_points['break_point'] = interpolation_break_points['break_point'] * \
            interpolation_break_points['generic_constraint_factor']
        interpolation_break_points['loss_segment'] = interpolation_break_points['loss_segment'] * \
            interpolation_break_points['generic_constraint_factor']
        interpolation_break_points = interpolation_break_points.drop('generic_constraint_factor', axis=1)

        loss_functions = pd.merge(interconnectors.loc[:, ['interconnector', 'link', 'generic_constraint_factor']],
                                  loss_functions, on='interconnector')

        def loss_function_adjuster(loss_function, generic_constraint_factor):
            def wrapper(flow):
                return loss_function(flow * generic_constraint_factor)
            return wrapper

        loss_functions['loss_function'] = \
            loss_functions.apply(lambda x: loss_function_adjuster(x['loss_function'], x['generic_constraint_factor']),
                                 axis=1)

        loss_functions = loss_functions.drop('generic_constraint_factor', axis=1)
        
        return loss_functions, interpolation_break_points

    def get_interconnector_definitions(self):
        """Returns inputs in the format needed to create interconnectors in the SpotMarket class.

        Examples
        --------

        >>> inputs_loader = _test_setup()

        >>> interconnector_data = InterconnectorData(inputs_loader)

        >>> interconnector_data.get_interconnector_definitions()
          interconnector from_region to_region     min     max       link  from_region_loss_factor  to_region_loss_factor  generic_constraint_factor
        0           V-SA        VIC1       SA1  -850.0   950.0       V-SA                   1.0000                 1.0000                          1
        1      N-Q-MNSP1        NSW1      QLD1  -264.0   264.0  N-Q-MNSP1                   1.0000                 1.0000                          1
        2      NSW1-QLD1        NSW1      QLD1 -1659.0  1229.0  NSW1-QLD1                   1.0000                 1.0000                          1
        3      V-S-MNSP1        VIC1       SA1  -270.0   270.0  V-S-MNSP1                   1.0000                 1.0000                          1
        5      VIC1-NSW1        VIC1      NSW1 -2299.0  2399.0  VIC1-NSW1                   1.0000                 1.0000                          1
        0      T-V-MNSP1        TAS1      VIC1     0.0   478.0    BLNKTAS                   1.0000                 0.9839                          1
        1      T-V-MNSP1        VIC1      TAS1     0.0   478.0    BLNKVIC                   0.9839                 1.0000                         -1

        Returns
        -------
        pd.DataFrame

            ========================  ================================
            Columns:                  Description:
            interconnector            unique identifier of a interconnector, \n
                                      (as `str`)
            to_region                 the region that receives power \n
                                      when flow is in the positive \n
                                      direction, (as `str`)
            from_region               the region that power is drawn \n
                                      from when flow is in the \n
                                      positive direction, (as `str`)
            max                       the maximum power flow on the \n
                                      interconnector, in MW (as `np.float64`)
            min                       the minimum power flow on the \n
                                      interconnector, if power can flow \n
                                      neative direction then this will be \n
                                      negative, in MW (as `np.float64`)
            from_region_loss_factor   the loss factor between the from \n
                                      end of the interconnector and the \n
                                      regional reference node, \n
                                      (as `np.float`)
            to_region_loss_factor     the loss factor between the to \n
                                      end of the interconnector and \n
                                      the regional reference node, \n
                                      (as `np.float`)
            ========================  ================================

        """
        regulated_interconnectors_series = \
            self.INTERCONNECTORCONSTRAINT[self.INTERCONNECTORCONSTRAINT['ICTYPE'] == 'REGULATED'].loc[:, 'INTERCONNECTORID']
        regulated_interconnectors = \
            self.interconnectors[self.interconnectors['interconnector'].isin(regulated_interconnectors_series)].copy()

        regulated_interconnectors['link'] = regulated_interconnectors['interconnector']
        regulated_interconnectors['from_region_loss_factor'] = 1.0
        regulated_interconnectors['to_region_loss_factor'] = 1.0
        regulated_interconnectors['generic_constraint_factor'] = 1

        market_interconnectors = self._get_market_interconnector_links()

        interconnectors = pd.concat([regulated_interconnectors, market_interconnectors])

        interconnectors['generic_constraint_factor'] = interconnectors['generic_constraint_factor'].astype(np.int64)

        return interconnectors

    def _get_market_interconnector_links(self):
        mnsp_bids = self.raw_input_loader.get_market_interconnector_link_bid_availability()
        MNSP_INTERCONNECTOR = self.raw_input_loader.get_market_interconnectors()
        mnsp_transmission_loss_factors = _format_mnsp_transmission_loss_factors(MNSP_INTERCONNECTOR,
                                                                                self.INTERCONNECTORCONSTRAINT)
        mnsp_transmission_loss_factors = \
            pd.merge(mnsp_transmission_loss_factors, mnsp_bids, on=['interconnector', 'to_region'])
        mnsp_transmission_loss_factors['max'] = np.where(~mnsp_transmission_loss_factors['availability'].isna(),
                                                         mnsp_transmission_loss_factors['availability'],
                                                         mnsp_transmission_loss_factors['max'])
        return mnsp_transmission_loss_factors.drop(columns=['availability'])


def _format_interconnector_definitions(INTERCONNECTOR, INTERCONNECTORCONSTRAINT):
    """
    Examples
    --------

    >>> INTERCONNECTOR = pd.DataFrame({
    ... 'INTERCONNECTORID': ['X', 'Y'],
    ... 'REGIONFROM': ['NSW', 'VIC'],
    ... 'REGIONTO': ['QLD', 'SA']})

    >>> INTERCONNECTORCONSTRAINT = pd.DataFrame({
    ... 'INTERCONNECTORID': ['X', 'Y'],
    ... 'IMPORTLIMIT': [100.0, 900.0],
    ... 'EXPORTLIMIT': [150.0, 800.0]})

    >>> interconnector_paramaters = _format_interconnector_definitions(INTERCONNECTOR, INTERCONNECTORCONSTRAINT)

    >>> print(interconnector_paramaters)
      interconnector from_region to_region    min    max
    0              X         NSW       QLD -100.0  150.0
    1              Y         VIC        SA -900.0  800.0

    """
    interconnector_directions = INTERCONNECTOR.loc[:, ['INTERCONNECTORID', 'REGIONFROM', 'REGIONTO']]
    interconnector_directions.columns = ['interconnector', 'from_region', 'to_region']
    interconnector_paramaters = INTERCONNECTORCONSTRAINT.loc[:, ['INTERCONNECTORID', 'IMPORTLIMIT', 'EXPORTLIMIT']]
    interconnector_paramaters.columns = ['interconnector', 'min', 'max']
    interconnector_paramaters['min'] = -1 * interconnector_paramaters['min']
    interconnectors = pd.merge(interconnector_directions, interconnector_paramaters, 'inner', on='interconnector')
    return interconnectors


def _format_mnsp_transmission_loss_factors(MNSP_INTERCONNECTOR, INTERCONNECTORCONSTRAINT):
    """
    Examples
    --------

    >>> MNSP_INTERCONNECTOR = pd.DataFrame({
    ...   'INTERCONNECTORID': ['A', 'A'],
    ...   'LINKID': ['A1', 'A2'],
    ...   'FROMREGION': ['C', 'X'],
    ...   'TOREGION': ['X', 'C'],
    ...   'FROM_REGION_TLF': [0.0, 0.75],
    ...   'TO_REGION_TLF': [0.75, 0.0],
    ...   'LHSFACTOR': [1.0, 0.9],
    ...   'MAXCAPACITY': [100.0, 200.0]})

    >>> INTERCONNECTORCONSTRAINT = pd.DataFrame({
    ...   'INTERCONNECTORID': ['A', 'B'],
    ...   'ICTYPE': ['MNSP', 'REGULATED']})

    >>> mnsp_transmission_loss_factors = _format_mnsp_transmission_loss_factors(MNSP_INTERCONNECTOR,
    ...   INTERCONNECTORCONSTRAINT)

    >>> print(mnsp_transmission_loss_factors.loc[:, ['interconnector', 'link', 'from_region', 'to_region']])
      interconnector link from_region to_region
    0              A   A1           C         X
    1              A   A2           X         C

    >>> print(mnsp_transmission_loss_factors.loc[:, ['interconnector', 'link', 'from_region_loss_factor',
    ...   'to_region_loss_factor']])
      interconnector link  from_region_loss_factor  to_region_loss_factor
    0              A   A1                     0.00                   0.75
    1              A   A2                     0.75                   0.00

    """
    INTERCONNECTORCONSTRAINT = INTERCONNECTORCONSTRAINT[INTERCONNECTORCONSTRAINT['ICTYPE'] == 'MNSP']
    MNSP_INTERCONNECTOR = pd.merge(MNSP_INTERCONNECTOR, INTERCONNECTORCONSTRAINT, on=['INTERCONNECTORID'])
    MNSP_INTERCONNECTOR = MNSP_INTERCONNECTOR.loc[:, ['INTERCONNECTORID', 'LINKID', 'FROM_REGION_TLF', 'TO_REGION_TLF',
                                                      'FROMREGION', 'TOREGION', 'LHSFACTOR', 'MAXCAPACITY']]

    mnsp_transmission_loss_factors = MNSP_INTERCONNECTOR.rename(columns={
        'INTERCONNECTORID': 'interconnector', 'LINKID': 'link', 'FROM_REGION_TLF': 'from_region_loss_factor',
        'TO_REGION_TLF': 'to_region_loss_factor', 'FROMREGION': 'from_region', 'TOREGION': 'to_region',
        'LHSFACTOR': 'generic_constraint_factor', 'MAXCAPACITY': 'max'
    })
    mnsp_transmission_loss_factors['min'] = 0.0
    return mnsp_transmission_loss_factors


def create_loss_functions(interconnector_coefficients, demand_coefficients, demand):
    """Creates a loss function for each interconnector.

    Transforms the dynamic demand dependendent interconnector loss functions into functions that only depend on
    interconnector flow. i.e takes the function f and creates g by pre-calculating the demand dependent terms.

        f(inter_flow, flow_coefficient, nsw_demand, nsw_coefficient, qld_demand, qld_coefficient) = inter_losses

    becomes

        g(inter_flow) = inter_losses

    The mathematics of the demand dependent loss functions is described in the
    :download:`Marginal Loss Factors documentation section 3 to 5  <../../docs/pdfs/Marginal Loss Factors for the 2020-21 Financial year.pdf>`.

    Examples
    --------
    >>> import pandas as pd

    Some arbitrary regional demands.

    >>> demand = pd.DataFrame({
    ...   'region': ['VIC1', 'NSW1', 'QLD1', 'SA1'],
    ...   'loss_function_demand': [6000.0 , 7000.0, 5000.0, 3000.0]})

    Loss model details from 2020 Jan NEM web LOSSFACTORMODEL file

    >>> demand_coefficients = pd.DataFrame({
    ...   'interconnector': ['NSW1-QLD1', 'NSW1-QLD1', 'VIC1-NSW1',
    ...                      'VIC1-NSW1', 'VIC1-NSW1'],
    ...   'region': ['NSW1', 'QLD1', 'NSW1', 'VIC1', 'SA1'],
    ...   'demand_coefficient': [-0.00000035146, 0.000010044,
    ...                           0.000021734, -0.000031523,
    ...                          -0.000065967]})

    Loss model details from 2020 Jan NEM web INTERCONNECTORCONSTRAINT file

    >>> interconnector_coefficients = pd.DataFrame({
    ...   'interconnector': ['NSW1-QLD1', 'VIC1-NSW1'],
    ...   'loss_constant': [0.9529, 1.0657],
    ...   'flow_coefficient': [0.00019617, 0.00017027],
    ...   'from_region_loss_share': [0.5, 0.5]})

    Create the loss functions

    >>> loss_functions = create_loss_functions(interconnector_coefficients,
    ...                                        demand_coefficients, demand)

    Lets use one of the loss functions, first get the loss function of VIC1-NSW1 and call it g

    >>> g = loss_functions[loss_functions['interconnector'] == 'VIC1-NSW1']['loss_function'].iloc[0]

    Calculate the losses at 600 MW flow

    >>> print(g(600.0))
    -70.87199999999996

    Now for NSW1-QLD1

    >>> h = loss_functions[loss_functions['interconnector'] == 'NSW1-QLD1']['loss_function'].iloc[0]

    >>> print(h(600.0))
    35.70646799999993

    Parameters
    ----------
    interconnector_coefficients : pd.DataFrame

        ======================  ======================================
        Columns:                Description:
        interconnector          unique identifier of a interconnector, \n
                                (as `str`)
        loss_constant           the constant term in the interconnector \n
                                loss factor equation, (as `np.float64`)
        flow_coefficient        the coefficient of the interconnector \n
                                flow variable in the loss factor equation \n
                                (as `np.float64`)
        from_region_loss_share  the proportion of loss attribute to the \n
                                from region, remainer are attributed to \n
                                the to region, (as `np.float64`)
        ======================  ======================================

    demand_coefficients : pd.DataFrame

        ==================  ==========================================
        Columns:            Description:
        interconnector      unique identifier of a interconnector, \n
                            (as `str`)
        region              the market region whose demand the coefficient \n
                            applies too (as `str`)
        demand_coefficient  the coefficient of regional demand variable \n
                            in the loss factor equation, (as `np.float64`)
        ==================  ==========================================

    demand : pd.DataFrame

        ====================  ========================================
        Columns:              Description:
        region                unique identifier of a region, (as `str`)
        loss_function_demand  the estimated regional demand, as calculated \n
                              by initial supply + demand forecast, \n
                              in MW (as `np.float64`)
        ====================  ========================================

    Returns
    -------
    pd.DataFrame

        loss_functions

        ================  ============================================
        Columns:          Description:
        interconnector    unique identifier of a interconnector, (as `str`)
        loss_function     a `function` object that takes interconnector \n
                          flow (as `float`) an input and returns \n
                          interconnector losses (as `float`).
        ================  ============================================
    """

    demand_loss_factor_offset = pd.merge(demand_coefficients, demand, 'inner', on=['region'])
    demand_loss_factor_offset['offset'] = demand_loss_factor_offset['loss_function_demand'] * \
        demand_loss_factor_offset['demand_coefficient']
    demand_loss_factor_offset = demand_loss_factor_offset.groupby('interconnector', as_index=False)['offset'].sum()
    loss_functions = pd.merge(interconnector_coefficients, demand_loss_factor_offset, 'left', on=['interconnector'])
    loss_functions['loss_constant'] = loss_functions['loss_constant'] + loss_functions['offset'].fillna(0)
    loss_functions['loss_function'] = \
        loss_functions.apply(lambda x: _create_function(x['loss_constant'], x['flow_coefficient']), axis=1)
    return loss_functions.loc[:, ['interconnector', 'loss_function', 'from_region_loss_share']]


def _create_function(constant, flow_coefficient):
    def loss_function(flow):
        return (constant - 1) * flow + (flow_coefficient / 2) * flow ** 2

    return loss_function
